import logging

import pandas as pd

import spherpro.bromodules.filter_base as filter_base
import spherpro.bromodules.filter_objectfilters as custfilter
import spherpro.configuration as conf
import spherpro.db as db
import spherpro.library as lib
from spherpro.bromodules.helpers_varia import HelperDb


class FilterMembership(filter_base.BaseFilter):
    def __init__(self, bro):
        super().__init__(bro)
        self.filter_custom = custfilter.ObjectFilterLib(bro)
        self.defaults = self.data.conf[conf.QUERY_DEFAULTS]
        self.helperdb = HelperDb(bro)

    def add_issmall(
        self, minpix=10, name=None, measid_area=None, object_type=None, drop=True
    ):
        if name is None:
            name = "is-small"
        obj_def = self.defaults[conf.OBJECT_DEFAULTS]
        measfilts = self.bro.filters.measurements

        # TODO: move to config
        if measid_area is None:
            fil = measfilts.get_measmeta_filter_statements(
                channel_names=[obj_def[conf.DEFAULT_CHANNEL_NAME]],
                stack_names=[obj_def[conf.DEFAULT_STACK_NAME]],
                measurement_names=["Area"],
                measurement_types=["AreaShape"],
            )
            measid_area = (
                self.data.get_measmeta_query()
                .filter(fil)
                .with_entities(db.measurements.measurement_id)
            ).one()[0]
        q_meas = (
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_id == measid_area)
            .add_columns(db.ref_stacks.scale)
        )

        q_obj = self.data.get_objectmeta_query()

        if object_type is not None:
            q_obj = q_obj.filter(db.objects.object_type == object_type)

        dat_meas = self.doquery(q_meas)
        dat_obj = self.doquery(q_obj)

        dat = self.bro.io.objmeasurements.get_measurements(dat_obj, dat_meas)
        self.bro.io.objmeasurements.scale_anndata(dat)

        dat_filter = pd.DataFrame(
            {
                db.objects.object_id.key: map(int, dat.obs.index),
                db.object_filters.filter_value.key: dat[:, str(measid_area)].X.squeeze()
                < minpix,
            }
        )
        self.filter_custom.write_filter_to_db(dat_filter, name, drop)
        return dat_filter

    def add_ismaincomponent(
        self, name=None, drop=True, relation="Neighbors", object_type="cell"
    ):
        if name is None:
            name = "is-maincomponent"
        # TODO: move to config
        dat_nb = self.helperdb.get_nb_dat(relation, obj_type=object_type)
        dat_obj = self.bro.doquery(
            self.session.query(db.objects.object_id, db.objects.image_id)
            .join(db.valid_objects)
            .filter(db.objects.object_type == object_type)
        )
        largest_obj = (
            dat_nb.merge(
                dat_obj,
                left_on=db.object_relations.object_id_parent.key,
                right_on=db.objects.object_id.key,
            )
            .groupby(db.images.image_id.key)
            .apply(lib.get_largest_commponent_objs)
        )
        dat_obj[db.object_filters.filter_value.key] = dat_obj[
            db.objects.object_id.key
        ].isin(largest_obj)
        self.filter_custom.write_filter_to_db(dat_obj, name, drop)
        return dat_obj

    def add_issphere(self, minfrac=0.6, name=None, drop=True, object_type="cell"):
        if name is None:
            name = "is-sphere"
        # TODO: move to config
        col_issphere = "is-sphere"
        col_isother = "is-other"
        col_isbg = "is-bg"
        col_measure = "MeanIntensity"
        col_stack = "BinStack"
        outcol_issphere = "is-sphere"
        non_zero_offset = 1 / 2 ** 20
        dat_meas = self.doquery(
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_name == col_measure)
            .filter(db.stacks.stack_name == col_stack)
            .filter(
                db.ref_planes.channel_name.in_([col_isother, col_issphere, col_isbg])
            )
            .add_columns(db.ref_stacks.scale, db.ref_planes.channel_name)
        )
        dat_obj = self.doquery(
            self.data.get_objectmeta_query().filter(
                db.objects.object_type == object_type
            )
        )
        logging.debug(f"{dat_obj.shape}")

        dat = self.bro.io.objmeasurements.get_measurements(dat_obj, dat_meas)
        logging.debug(f"{dat.shape}")
        self.bro.io.objmeasurements.scale_anndata(dat)

        fil_issphere, fil_isbg, fil_isother = (
            dat.var[db.ref_planes.channel_name.key] == c
            for c in (col_issphere, col_isbg, col_isother)
        )
        dat_filter = pd.DataFrame(
            {
                db.objects.object_id.key: map(int, dat.obs.index),
                db.object_filters.filter_value.key: (
                    (
                        (dat.X[:, fil_issphere] + non_zero_offset)
                        / (
                            dat.X[:, fil_isother]
                            + dat.X[:, fil_isbg]
                            + dat.X[:, fil_issphere]
                            + non_zero_offset
                        )
                    )
                    > minfrac
                )
                .astype(int)
                .flatten(),
            }
        )
        self.filter_custom.write_filter_to_db(dat_filter, name, drop)
        return dat_filter

    def add_isambiguous(self, distother=-10, name=None, drop=True, object_type="cell"):
        if name is None:
            name = "is-ambiguous"
        # TODO: move to config
        col_measure = "MeanIntensity"
        col_stack = "DistStack"
        col_distother = "dist-other"
        measid = (
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_name == col_measure)
            .filter(db.stacks.stack_name == col_stack)
            .filter(db.ref_planes.channel_name == col_distother)
            .with_entities(db.measurements.measurement_id)
            .one()[0]
        )

        q_meas = (
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_id == measid)
            .add_columns(db.ref_stacks.scale)
        )
        q_obj = self.data.get_objectmeta_query()

        if object_type is not None:
            q_obj = q_obj.filter(db.objects.object_type == object_type)

        dat_meas = self.doquery(q_meas)
        dat_obj = self.doquery(q_obj)

        dat = self.bro.io.objmeasurements.get_measurements(dat_obj, dat_meas)
        self.bro.io.objmeasurements.scale_anndata(dat)

        d = dat[:, str(measid)].X.squeeze()
        dat_filter = pd.DataFrame(
            {
                db.objects.object_id.key: map(int, dat.obs.index),
                db.object_filters.filter_value.key: (d > distother)
                & (d < (2 ** 16 - 2)).astype(int),
            }
        )
        self.filter_custom.write_filter_to_db(dat_filter, name, drop)
        return dat_filter

    def add_isnotborder(self, borderdist=5, name=None, drop=True, object_type="cell"):
        if name is None:
            name = "is-notborder"

        # TODO: move to config
        col_measure = "MinIntensity"
        col_stack = "DistStack"
        col_distsphere = "dist-sphere"

        measid = (
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_name == col_measure)
            .filter(db.stacks.stack_name == col_stack)
            .filter(db.ref_planes.channel_name == col_distsphere)
            .with_entities(db.measurements.measurement_id)
            .one()[0]
        )
        q_meas = (
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_id == measid)
            .add_columns(db.ref_stacks.scale)
        )
        q_obj = self.data.get_objectmeta_query()

        if object_type is not None:
            q_obj = q_obj.filter(db.objects.object_type == object_type)

        dat_meas = self.doquery(q_meas)
        dat_obj = self.doquery(q_obj)

        dat = self.bro.io.objmeasurements.get_measurements(dat_obj, dat_meas)
        self.bro.io.objmeasurements.scale_anndata(dat)

        d = dat[:, str(measid)].X.squeeze()
        dat_filter = pd.DataFrame(
            {
                db.objects.object_id.key: map(int, dat.obs.index),
                db.object_filters.filter_value.key: (d > borderdist)
                & (d < (2 ** 16 - 2)).astype(int),
            }
        )
        self.filter_custom.write_filter_to_db(dat_filter, name, drop)
        return dat_filter



import numpy as np

import spherpro.db as db

DEFAULT_RELATION = "Neighbors"
VALUE = db.object_measurements.value.key
MEAS_ID = db.measurements.measurement_id.key
CHILD_ID = db.object_relations.object_id_child.key
PARENT_ID = db.object_relations.object_id_parent.key
OBJ_ID = db.objects.object_id.key
OBJ_TYPE = db.objects.object_type.key
OLD_ID = "oldid"


class AggregateNeightbours(object):
    """docstring for CalculateDistRim."""

    def __init__(self, bro):
        self.bro = bro
        self.session = self.bro.data.main_session
        self.data = self.bro.data
        self.mm = self.bro.processing.measurement_maker
        self._register_measurement_name = self.mm.register_measurement_name
        self._register_measurement_type = self.mm.register_measurement_type

    def add_nb_measurement(
        self,
        nb_meas_prefix,
        nb_agg_fkt,
        nb_relationtype=None,
        object_type=None,
        measurement_name=None,
        stack_name=None,
        plane_id=None,
        measurement_type=None,
        filter_query=None,
        filter_statement=None,
        image_id=None,
        valid_objects=True,
    ):
        if nb_relationtype is None:
            nb_relationtype = DEFAULT_RELATION
        if filter_statement is not None:
            raise ("filter_statement not implemented yet")

        nbfil = filter_query
        nb_dic_dat = self.get_nb_dat(
            nb_relationtype,
            obj_type=object_type,
            fil_query=nbfil,
            valid_objects=valid_objects,
        )
        dat = self._get_data(
            object_type,
            measurement_name,
            stack_name,
            plane_id,
            measurement_type,
            filter_query,
            filter_statement,
            image_id,
            valid_objects=valid_objects,
        )

        fil = (
            (nb_dic_dat[CHILD_ID].isin(dat[OBJ_ID]))
            & (nb_dic_dat[PARENT_ID].isin(dat[OBJ_ID]))
            & (nb_dic_dat[PARENT_ID] != nb_dic_dat[CHILD_ID])
        )
        nb_dic_dat = nb_dic_dat.loc[fil, :]
        nb_dict = self._gen_nb_dict(nb_dic_dat)
        dat = dat.loc[dat[OBJ_ID].isin(nb_dic_dat[PARENT_ID]), :]
        nb_dat = self.agg_data(dat, nb_dict, nb_agg_fkt)
        old_ids = [int(i) for i in dat[MEAS_ID].unique()]
        id_dict = self.update_measurement_ids(old_ids, nb_meas_prefix)
        nb_dat[MEAS_ID] = nb_dat[MEAS_ID].replace(id_dict)
        self.mm.add_object_measurements(nb_dat, drop_all_old=True)

    def agg_data(self, data, nb_dict, fkt):
        tdat = data.pivot_table(index=[OBJ_ID, OBJ_TYPE], columns=MEAS_ID, values=VALUE)
        nb_dat = tdat.apply(
            self._agg_nb_val,
            axis=1,
            nbdict=nb_dict,
            data=tdat.reset_index(OBJ_TYPE, drop=True),
            fkt=fkt,
        )
        nb_dat = nb_dat.stack()
        nb_dat.name = VALUE
        nb_dat = nb_dat.reset_index(drop=False)
        return nb_dat

    def get_nb_dat(
        self, relationtype_name, obj_type=None, fil_query=None, valid_objects=True
    ):
        return self.bro.helpers.dbhelp.get_nb_dat(
            relationtype_name, obj_type, fil_query, valid_obj_only=valid_objects
        )

    def _get_data(
        self,
        object_type=None,
        measurement_name=None,
        stack_name=None,
        plane_id=None,
        measurement_type=None,
        filter_query=None,
        filter_statement=None,
        image_id=None,
        valid_objects=True,
    ):
        q_obj = self.data.get_objectmeta_query(valid_objects=valid_objects)
        q_meas = self.data.get_measmeta_query()

        if object_type is not None:
            q_obj = q_obj.filter(db.objects.object_type == object_type)
        if measurement_name is not None:
            q_meas = q_meas.filter(db.measurements.measurement_name == measurement_name)
        if measurement_type is not None:
            q_meas = q_meas.filter(db.measurements.measurement_type == measurement_type)
        if stack_name is not None:
            q_meas = q_meas.filter(db.stacks.stack_name == stack_name)
        if filter_query is not None:
            q_obj = q_obj.filter(db.objects.object_id == filter_query.c.object_id)
        if filter_statement is not None:
            q_obj = q_obj.filter(filter_statement)
        if plane_id is not None:
            q_meas = q_meas.filter(db.planes.plane_id == plane_id)
        if image_id is not None:
            q_obj = q_obj.filter(db.images.image_id == image_id)
        adat = self.bro.io.objmeasurements.get_measurements(q_obj=q_obj, q_meas=q_meas)
        dat = self.bro.io.objmeasurements.convert_anndata_legacy(adat)
        return dat

    @staticmethod
    def _agg_nb_val(row, data, nbdict, fkt):
        objid = row.name[0]
        y = data.loc[list(nbdict[objid]), :].apply(fkt, axis=0, raw=True)
        return y

    @staticmethod
    def _gen_nb_dict(coldat):
        """
        Converts relationship lists to dict
        """
        first_obj = coldat[PARENT_ID].values
        second_obj = coldat[CHILD_ID].values
        nb_dict = dict((obj, set()) for obj in np.unique(first_obj))
        for f, s in zip(first_obj, second_obj):
            nb_dict[f].add(s)
        return nb_dict

    def update_measurement_ids(self, old_ids, meas_name_prefix):
        measure_meta = self.bro.doquery(
            self.session.query(db.measurements).filter(
                db.measurements.measurement_id.in_(old_ids)
            )
        )
        dic_meas_name = dict()
        for m in measure_meta[db.measurements.measurement_name.key].unique():
            m_new = meas_name_prefix + m
            dic_meas_name.update({m: m_new})
            self._register_measurement_name(m_new)
        measure_meta[db.measurements.measurement_name.key] = measure_meta[
            db.measurements.measurement_name.key
        ].replace(dic_meas_name)
        measure_meta = measure_meta.rename(columns={MEAS_ID: OLD_ID})
        measure_meta = self.mm.register_measurements(measure_meta)
        id_dict = {
            old: new
            for old, new in zip(measure_meta[OLD_ID], measure_meta[MEAS_ID])
            if old != new
        }
        return id_dict



"""
A class to generate and add filters to the filter table.
"""
import pandas as pd
import sqlalchemy as sa

import spherpro.bromodules.filter_base as filter_base
import spherpro.db as db

# TODO: move to default configuration?
FILTERSTACKNAME = "FilterStack"
FILTERTYPENAME = "filter"


class ObjectFilterLib(filter_base.BaseFilter):
    def __init__(self, bro):
        super().__init__(bro)

    def add_filtername(self, filtername):
        """
        Adds a filtername to the objectfilter
        Args:
            filtername: a string
        """
        fil_id = (
            self.data.main_session.query(db.object_filter_names.object_filter_id)
            .filter(db.object_filter_names.object_filter_name == filtername)
            .scalar()
        )
        if fil_id is None:
            new_id = self.data._query_new_ids(
                db.object_filter_names.object_filter_id, 1
            )
            new_id = list(new_id)
            fil_id = new_id[0]
            dat = pd.DataFrame(
                {
                    db.object_filter_names.object_filter_name.key: [filtername],
                    db.object_filter_names.object_filter_id.key: new_id,
                }
            )
            self.data._add_generic_tuple(dat, db.object_filter_names)

        fil_id = int(fil_id)

        return fil_id

    def write_filter_to_db(self, filterdata, filtername, drop=True, replace=True):
        """
        Writes a dataframe containing Filterdata to the DB.
        Args:
            filterdata: DataFrame containing the filterdata. Needs to contain a column filter_value
                and object_id
            filtername: String stating the Filtername
        """
        filterdata = filterdata.loc[
            :, [db.objects.object_id.key, db.object_filters.filter_value.key]
        ]
        filterdata.loc[
            :, db.object_filter_names.object_filter_id.key
        ] = self.add_filtername(filtername)

        if drop:
            self.delete_filter_by_name(filtername)
        filterdata = filterdata.dropna()
        filterdata[db.object_filters.filter_value.key] = filterdata[
            db.object_filters.filter_value.key
        ].astype(int)
        self.data._bulkinsert(filterdata, db.object_filters)

    def delete_filter_by_name(self, filtername):
        fid = (
            self.session.query(db.object_filter_names.object_filter_id).filter(
                db.object_filter_names.object_filter_name == filtername
            )
        ).one()[0]
        (
            self.session.query(db.object_filters).filter(
                db.object_filters.object_filter_id == fid
            )
        ).delete()
        self.session.commit()

    def get_combined_filterquery(self, object_filters):
        """
        Get a filter query for the requested filters:
        Args:
            object_filters: list of format [(filtername1, filtervalue1),
                                            (filtername2, filtervalue2), ... ]
            image_filters: list of same format as object_filters
        returns: a subquery that can be joined to another query
        """

        subquerys = [
            self.data.main_session.query(db.object_filters.object_id)
            .join(db.object_filter_names)
            .filter(db.object_filter_names.object_filter_name == filname)
            .filter(db.object_filters.filter_value == int(filval))
            .subquery(filname + str(filval))
            for filname, filval in object_filters
        ]
        query = self.data.main_session.query(db.objects.object_id)
        for sq in subquerys:
            query = query.join(sq, db.objects.object_id == sq.c.object_id)
        return query.subquery()

    def get_combined_filterstatement(self, object_filters):
        """
        Get a filter statement for the requested filters:
        Args:
            object_filters: list of format [(filtername1, filtervalue1),
                                            (filtername2, filtervalue2), ... ]
        returns: a subquery that can be joined to another query
        """
        subquery = self.get_combined_filterquery(object_filters)
        fil = sa.and_(db.objects.object_id == subquery.c.object_id)
        return fil

    def delete_by_filter(self, filter_statement):
        """
        Deletes objects from the 'valid_objects' list based
        on an object filter statement.

        Args:
            filter_statement: an object filter statement

        Returns:

        """
        stmt = self.session.query(db.valid_objects).filter(
            db.valid_objects.object_id.in_(filter_statement)
        )
        stmt.delete(synchronize_session="fetch")
        self.session.commit()



"""
A class to generate handle the loading of the mask specified in the database.
"""
import functools
import os

import numpy as np
import tifffile as tif

import spherpro.bromodules.io_base as io_base
import spherpro.configuration as conf
import spherpro.db as db

max_cache = 384


class IoMasks(io_base.BaseIo):
    def __init__(self, bro):
        super().__init__(bro)
        cpconf = self.data.conf[conf.CPOUTPUT]
        self.basedir = cpconf[conf.IMAGES_CSV][conf.MASK_DIR]
        if self.basedir is None:
            self.basedir = self.data.conf[conf.CP_DIR]
        else:
            self.basedir = self.basedir.format(
                **{conf.CP_DIR: self.data.conf[conf.CP_DIR]}
            )

        self._dat_masks = None

    @functools.lru_cache(maxsize=max_cache)
    def get_mask(self, image_id: int, object_type: str) -> np.ndarray:
        """
        Retrieves masks based on an image_id
        Args:
            image_id: Image number
            object_type: Object type
        Returns:
            mask_array: numpy array with the mask labels as integer image
        """
        fn = (
            self.bro.session.query(db.masks.mask_filename)
            .filter(db.masks.image_id == image_id)
            .filter(db.masks.object_type == object_type)
        ).one()[0]
        return tif.imread(os.path.join(self.basedir, fn))

    def clear_caches(self):
        self.get_mask.cache_clear()

    @property
    def dat_masks(self):
        if self._dat_masks is None:
            q = self.bro.session.query(db.masks)
            self._dat_masks = self.bro.doquery(q)
        return self._dat_masks



"""
A class to generate filter queries from measurements
and save them into the database.
"""
from itertools import chain, repeat

import numpy as np
import sqlalchemy as sa

import spherpro.bromodules.filter_base as filter_base
import spherpro.db as db


class FilterMeasurements(filter_base.BaseFilter):
    def __init__(self, bro):
        super().__init__(bro)
        self.measure_idx = [  # idx_name, default
            (db.objects.object_type.key, "cell"),
            (db.ref_planes.channel_name.key, None),
            (db.stacks.stack_name.key, "FullStack"),
            (db.measurement_names.measurement_name.key, "MeanIntensity"),
            (db.measurement_types.measurement_type.key, None),
        ]
        self.get_filter_vector = get_filter_vector

    def get_measmeta_filter_statements(
        self, channel_names, stack_names, measurement_names, measurement_types
    ):
        """
        Generates a filter expression to filter measurements by,
        stack_names, measurement names and measurement types.

        Input:
            channel_names: list of channel names
            stack_names: list ofstack_names
            measurement_names: list of measurement_names
            measurement_types: list of measurement measurement_types
        Returns:
            A dataframes with the selected measurements
        """
        constraint_columns = [
            db.ref_planes.channel_name,
            db.stacks.stack_name,
            db.measurements.measurement_name,
            db.measurement_types.measurement_type,
        ]

        value_lists = [channel_names, stack_names, measurement_names, measurement_types]
        measure_filter = combine_constraints(
            constraint_columns, value_lists=value_lists
        )
        return measure_filter

    def get_objectmeta_filter_statements(self, object_types):
        """
        Generates a filter expression to filter measurements by,
        stack_names, measurement names and measurement types.

        Input:
            channel_names: list of channel names
            stack_names: list of stack_names
            measurement_names: list of measurement_names
            measurement_types: list of measurement measurement_types
        Returns:
            A dataframes with the selected measurements
        """
        constraint_columns = [db.objects.object_type]

        value_lists = [object_types]
        measure_filter = combine_constraints(
            constraint_columns, value_lists=value_lists
        )
        return measure_filter

    def get_filter_data(self, dat_obj, filter_triplets):
        measids = set(f[0] for f in filter_triplets)
        anndat = self.bro.io.objmeasurements.get_measurements(dat_obj, measidx=measids)
        return anndat

    def measmeta_to_measid(
        self,
        channel_name=None,
        stack_name=None,
        measurement_name=None,
        measurement_type=None,
    ):
        fil = self.get_measmeta_filter_statements(
            channel_names=[channel_name],
            stack_names=[stack_name],
            measurement_names=[measurement_name],
            measurement_types=[measurement_type],
        )
        measid = (
            self.data.get_measmeta_query()
            .filter(fil)
            .with_entities(db.measurements.measurement_id)
            .all()
        )
        if len(measid) > 1:
            raise ValueError(
                f"Measurment not uniquely specified.\n {len(measid)} measurements found that match specification."
            )
        return measid[0][0]


def get_filter_vector(anndat, filter_triplets):
    barr = None
    for m, l, t in filter_triplets:
        d = anndat[:, str(m)].X
        tmp = l(d, t)
        if barr is None:
            barr = np.array(tmp)
        else:
            barr = barr & tmp
    return barr


def combine_constraints(columns, value_lists):
    """
    Combine constraints on columns in an sqlalchemy query.
    Args:
        columns: table columns to constrain on
        value_lists: lists of constraints:
            either contain elements
            OR tuples with multiple elements
            OR None to not constrain
    Returns:
        A filter statement encoding the constraints.
    """
    constraints = [
        sa.and_(
            *[
                c.in_(v) if isinstance(v, tuple) else c == v
                for c, v in zip_exact(columns, values)
                if v is not None
            ]
        )
        for values in zip_exact(*value_lists)
    ]
    if len(constraints) > 1:
        measure_filter = sa.or_(*constraints)
    else:
        measure_filter = constraints[0]
    return measure_filter


def zip_exact(*args):
    """
    A zip that asserts that all list have equal length
    """
    sentinel = object()
    iters = [chain(it, repeat(sentinel)) for it in args]
    for result in zip(*iters):
        if sentinel in result:
            if all(value == sentinel for value in result):
                return
            raise ValueError("sequences of different lengths")
        yield result



"""
A class to generate handle the loading of the mask specified in the database.
"""
import functools
import os
import re

import imctools.io.imcacquisition as imcacquisition
import imctools.io.ometiffparser as omepars
import numpy as np

import spherpro.bromodules.io_base as io_base
import spherpro.configuration as conf
import spherpro.db as db

max_cache = 384


class IoImc(io_base.BaseIo):
    def __init__(self, bro):
        super().__init__(bro)
        imgconf = self.data.conf[conf.CPOUTPUT][conf.IMAGES_CSV]

        self.ome_dirs = imgconf[conf.IMAGE_OME_FOLDER_DIRS]
        self.meta_re = re.compile(imgconf[conf.IMAGE_OME_META_REGEXP])
        self._ome_folddict = None

    @functools.lru_cache(maxsize=max_cache)
    def get_imc_acquisition(self, slideac_name, acid):
        """
        Retrieves masks based on a folder and acquisition_id
        Args:
            folder
        Returns:
            memmapped imcacquisition
        """
        fn_img = self.ome_folddict[slideac_name][acid]
        return omepars.OmetiffParser(fn_img).get_imc_acquisition()

    def get_imcimg(self, img_id):
        meta = self._get_imgmeta(img_id)
        slideac_name = meta[db.slideacs.slideac_name.key]
        acid = meta[db.acquisitions.acquisition_mcd_acid.key]
        imcac = self.get_imc_acquisition(slideac_name, acid)
        posx = int(meta[db.images.image_pos_x.key])
        posy = int(meta[db.images.image_pos_y.key])
        w = int(meta[db.images.image_shape_w.key])
        h = int(meta[db.images.image_shape_h.key])
        sl = np.s_[:, posx : (h + posx), posy : (w + posy)]
        cutac = imcacquisition.ImcAcquisition(
            imcac.image_ID,
            imcac.original_file,
            imcac.data[sl],
            imcac._channel_metals,
            imcac._channel_labels,
        )
        cutac.original_imcac = imcac
        return cutac

    def clear_caches(self):
        self.get_imc_acquisition.cache_clear()
        self._get_imgmeta.cache_clear()

    @functools.lru_cache(maxsize=max_cache)
    def _get_imgmeta(self, imgid):
        q = (
            self.data.main_session.query(
                db.images.image_id,
                db.images.image_pos_x,
                db.images.image_pos_y,
                db.images.image_shape_w,
                db.images.image_shape_h,
                db.acquisitions.acquisition_mcd_acid,
                db.slideacs.slideac_name,
            )
            .join(db.acquisitions, db.sites, db.slideacs)
            .filter(db.images.image_id == imgid)
        )
        r = {
            l: v
            for r in [q.one_or_none()]
            if r is not None
            for l, v in zip(r.keys(), r)
        }
        return r

    @property
    def ome_folddict(self):
        if self._ome_folddict is None:
            self._ome_folddict = {
                subfol: self._get_acs_from_fol(os.path.join(fol, subfol))
                for fol in self.ome_dirs
                for subfol in os.listdir(fol)
                if os.path.isdir(os.path.join(fol, subfol))
            }
        return self._ome_folddict

    def _get_acs_from_fol(self, fol):
        acdict = {}
        for fn in os.listdir(fol):
            m = self.meta_re.match(fn)
            if m is not None:
                acid = int(m.groupdict()[db.acquisitions.acquisition_mcd_acid.key])
                acdict[acid] = os.path.join(fol, fn)
        return acdict



import os
import re

import networkx as nx
import pandas as pd

import spherpro.configuration as conf
import spherpro.db as db


def fill_null(data, table):
    """
    Fills a Dataframe with less columns as the DB table with None
    """
    data_cols = data.columns
    table_cols = table.__table__.columns.keys()
    uniq = list(set(table_cols) - set(data_cols))
    for un in uniq:
        data[un] = None
    return data


def find_measurementmeta(stack_name, col_name, no_stack_str=None, no_plane_string=None):
    """
    finds the measurement meta information from a given string

    Args:
        stack_name: an iterable  containing a name for all known stacks.
            exp: [DistStack, BinStackr, FullStack]
        columns: a list of column names.

    Returns:
        Returns pandas.Series, with the following strings
        in this order:
        x | measurement type | measurement name | stack name | plane id
    """
    stackpattern = "(" + "|".join(stack_name) + ")"
    pre_pattern = "^([^_]*)_(.*)"
    post_pattern = "(.*)_" + stackpattern + "_(c\d+)"
    match = re.search(pre_pattern, col_name)
    if match != None:
        match = match.groups()
        mtype = match[0]
        rside = match[1]
    else:
        return pd.Series(["", "", "", "", ""])

    match = re.search(post_pattern, rside)
    if match != None:
        match = match.groups()
        name = match[0]
        stack = match[1]
        plane = match[2]
    else:
        name = rside
        stack = no_stack_str
        plane = no_plane_string

    return pd.Series([col_name, mtype, name, stack, plane])


def construct_in_clause_list(key_dict):
    """
    Construnct a list of IN clauses.
    Args:
        key_dict: a dictionary where key is the database column key and value
            is the values to filter for.
    Return:
        a list of constructed clauses

    Example:
        >>> out = construct_in_clause_list({'A': [1], 'B': ['c','d']})
        >>> sorted(out)
        ['A IN ("1")', 'B IN ("c","d")']

    """
    querylist = [
        k + ' IN ("' + '","'.join(map(str, values)) + '")'
        for k, values in key_dict.items()
    ]
    return querylist


def filter_and_rename_dict(indict, filterdict):
    """
    Renames the  keys of the input  dict using the filterdict.
    Keys not present in the filterdict will be removed.

    Example:
        >>> import pprint
        >>> outdict =filter_and_rename_dict({'a': 1, 'b':[2,2]}, {'a': 'c', 'b': 'd', 'e': 'f'})
        >>> pprint.pprint(outdict)
        {'c': 1, 'd': [2, 2]}

    """
    outdict = {
        filterdict[k]: v
        for k, v in indict.items()
        if ((v is not None) and (k in filterdict.keys()))
    }

    return outdict


def read_csv_from_config(configdict, base_dir=None):
    """
    Read the CSV from a configuration entry.
    """
    path = configdict[conf.PATH]
    sep = configdict[conf.SEP]
    if base_dir is not None:
        path = os.path.join(base_dir, path)
    dat = pd.read_csv(path, sep=sep)
    return dat


def map_group_re(x, re_str):
    """
    Maps a regular expression with matchgroups
    to a iterable (e.g. column in a dataframe) and returns the
    result as a data frame
    Args:
        x: iterable
        re_str: a regular expression string with matchgroups
    Return:
        A dataframe with column names being matchgroups

    """
    qre = re.compile(re_str)
    m_list = [
        pd.DataFrame.from_dict([m.groupdict() for m in qre.finditer(s)]) for s in x
    ]
    return pd.concat(m_list, ignore_index=True)


def get_largest_commponent_objs(dat, keys=None):
    """
    Get the nodes of the largest connected component of a
    graph.
    Args:
        dat: a dataframe representing the edgelist
        keys: list of source and target column names
    Return:
        A list of object ids
    """
    if keys is None:
        keys = [
            db.object_relations.object_id_parent.key,
            db.object_relations.object_id_child.key,
        ]
    g = nx.from_pandas_edgelist(dat[keys], source=keys[0], target=keys[1])
    gmax = max(nx.connected_components(g), key=len)
    return pd.Series((int(n) for n in gmax), name=db.objects.object_id.key)



import matplotlib.pyplot as plt
import numpy as np
import pycytools as pct
from matplotlib_scalebar.scalebar import ScaleBar

import spherpro.bromodules.plot_base as plot_base
import spherpro.db as db

"""
TODO:
This is work in progress!
"""


class PlotImage(plot_base.BasePlot):
    """
    A class to plot single images on an axes:
    - intensity image
    - heatmask image
    - contour image

    """

    def __init__(self, bro):
        super().__init__(bro)
        self.io_masks = self.bro.io.masks
        self.filter_measurements = self.bro.filters.measurements
        self.objmeasurements = self.bro.io.objmeasurements
        self.measure_idx = [  # idx_name, default
            (db.ref_planes.channel_name.key, None),
            (db.stacks.stack_name.key, "FullStackFiltered"),
            (db.measurement_names.measurement_name.key, "MeanIntensity"),
            (db.measurement_types.measurement_type.key, None),
        ]


def plot_contour(img_id, object_numbers, values, cmap):
    """
    Plots values
    Args:
        img_id:
        object_numbers:
        values:
        cmap:

    Returns:

    """
    raise NotImplementedError
    pass


def add_scalebar(
    ax, resolution=0.000001, location=4, color="white", pad=0.5, frameon=False, **kwargs
):
    scalebar = ScaleBar(
        resolution, location=location, color=color, pad=pad, frameon=frameon, **kwargs
    )  # 1 pixel = 0.2 meter
    ax.add_artist(scalebar)


def adapt_ax_clims(axs):
    caxs = [ax.images[0] for ax in axs if len(ax.images) > 0]
    clims = [cax.get_clim() for cax in caxs]
    clims = [c for c in clims if c != (True, True)]
    clim_all = [f(c) for f, c in zip([np.min, np.max], zip(*clims))]
    for cax in caxs:
        cax.set_clim(clim_all)


def map_img(mask, values, objectnr):
    timg = pct.library.map_series_on_mask(
        mask, dat_curimg[valuevar], label=dat_curimg[V.COL_OBJ_NR]
    )
    return timg


def add_cb_to_last(axs):
    caxs = [cax for ax in axs for cax in ax.images if ax.images is not None]
    plt.colorbar(caxs[-1], ax=list(axs)[-1])



import matplotlib.pyplot as plt
import matplotlib_scalebar.scalebar as scalebar
import numpy as np

import spherpro.bromodules.io_stackimage as io_stackimage
import spherpro.bromodules.plot_base as plot_base
import spherpro.db as db

LABEL_Y = "Condition ID number"
LABEL_X = "Image ID number"
PLT_TITLE = "All images from a single condition"

# TODO: take all defaults from the configuration!


class PlotConditionImages(plot_base.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)
        # make the dependency explicit
        self.heatmask = bro.plots.heatmask
        self.measurement_filters = bro.filters.measurements
        self.objectfilterlib = bro.filters.objectfilterlib
        self.imcimage = bro.io.imcimg
        self.stackimage = io_stackimage.IoStackImage(bro)
        self.get_target_by_channel = bro.helpers.dbhelp.get_target_by_channel

    def plot_hm_conditions(
        self,
        condition_name,
        channel_name,
        stack_name="FullStackFiltered",
        measurement_name="MeanIntensity",
        object_type="cell",
        minmax=(0, 1),
        transf=None,
    ):
        cond_list = self.get_cond_id_im_id(condition_name)
        im_dict = self.get_dict_imgs(
            cond_list, channel_name, stack_name, measurement_name, object_type
        )
        if transf is not None:
            for key, val in im_dict.items():
                im_dict[key] = transf(val)

        target = self.get_target_by_channel(channel_name)
        title = "condition: %s\nchannel: %s - %s" % (
            condition_name,
            channel_name,
            target,
        )

        fig, hm = self.plot_layout(cond_list, im_dict, title, minmax=minmax)

        return fig

    def plot_imc_conditions(
        self, condition_name, channel_name, minmax=(0, 1), transf=None
    ):

        cond_list = self.get_cond_id_im_id(condition_name)
        im_dict = self.get_dict_imc_imgs(cond_list, channel_name)
        if transf is not None:
            for key, val in im_dict.items():
                im_dict[key] = transf(val)

        target = self.get_target_by_channel(channel_name)
        title = "condition: %s\nchannel: %s - %s" % (
            condition_name,
            channel_name,
            target,
        )

        fig, hm = self.plot_layout(cond_list, im_dict, title, minmax=minmax)

        return fig

    def plot_stackimg_conditions(
        self, condition_name, channel_name, stack_name, minmax=(0, 1), transf=None
    ):
        cond_list = self.get_cond_id_im_id(condition_name)
        im_dict = self.get_dict_stack_imgs(cond_list, channel_name, stack_name)
        if transf is not None:
            for key, val in im_dict.items():
                im_dict[key] = transf(val)

        target = self.get_target_by_channel(channel_name)
        title = "condition: %s\nchannel: %s - %s" % (
            condition_name,
            channel_name,
            target,
        )

        fig, hm = self.plot_layout(cond_list, im_dict, title, minmax=minmax)

        return fig

    def plot_layout(
        self, cond_list, im_dict, title, pltfkt=None, minmax=(0, 1), crange=None
    ):

        if pltfkt is None:
            pltfkt = self.plot_im

        nrows = len(cond_list)
        ncols = max([len(c[1]) for c in cond_list])

        if crange is None:
            crange = self.get_crange(im_dict, minmax)

        cond_id, image_id = zip(*cond_list)

        shape = [(np.shape(i)) for i in im_dict.values()]
        x_shape = max(shape, key=lambda x: x[0])[0]
        y_shape = max(shape, key=lambda x: x[1])[1]

        fig, ax = plt.subplots(
            nrows, ncols, figsize=(2 * ncols + 2, 2 * nrows + 2), squeeze=True
        )
        if nrows == 1:
            ax = np.array([ax])
        if ncols == 1:
            ax = np.array([[a] for a in ax])

        for i, axrow in enumerate(ax):
            cond, images = cond_list[i]

            for j, a in enumerate(axrow):
                if j < len(images):
                    image = images[j]
                    img = im_dict[image]
                    cax = pltfkt(img, ax=a, crange=crange)
                    sb = scalebar.ScaleBar(
                        1, units="um", location=4, frameon=False, color="white"
                    )
                    a.add_artist(sb)
                    a.set_xticks([])
                    a.set_yticks([])
                    a.set_title("Im_id: %s" % str(image), size="small")
                    if j == 0:
                        a.set_ylabel(
                            "Cond_id: %s" % str(cond),
                            rotation=0,
                            size="small",
                            labelpad=39,
                        )

                else:
                    a.set_visible(False)

        plt.colorbar(cax.images[0], ax=ax.ravel().tolist())
        plt.suptitle(title)
        return fig, ax

    def plot_im(
        self, img, title=None, crange=None, ax=None, update_axrange=True, cmap=None
    ):
        cax = self.heatmask.do_heatplot(
            img=img,
            title=title,
            crange=crange,
            ax=ax,
            update_axrange=update_axrange,
            cmap=cmap,
            colorbar=False,
        )
        return cax

    @staticmethod
    def get_crange(img_dict, minmax=(0, 1)):
        vals = np.concatenate([v[np.isnan(v) == False] for v in img_dict.values()])
        crange = [
            np.percentile(vals, 100 * minmax[0]),
            np.percentile(vals, 100 * minmax[1]),
        ]
        return crange

    def get_dict_imgs(
        self, cond_list, channel_name, stack_name, measurement_name, object_type
    ):
        img_ids = [int(i) for c, imgs in cond_list for i in imgs]
        dat_obj = self.heatmask.get_heatmask_data(
            {
                db.stacks.stack_name.key: stack_name,
                db.measurements.measurement_name.key: measurement_name,
                db.ref_planes.channel_name.key: channel_name,
            },
            image_ids=img_ids,
            object_type=object_type,
        )

        imgs = {
            img_id: self.heatmask.assemble_heatmap_image(
                dat_obj.query(f"{db.images.image_id.key} == {img_id}")
            )
            for c, imgs in cond_list
            for img_id in imgs
        }
        return imgs

    def get_dict_imc_imgs(self, cond_list, channel_name):
        imac = {
            img: self.imcimage.get_imcimg(int(img))
            for c, imgs in cond_list
            for img in imgs
        }
        for key, val in imac.items():
            imac[key] = val.get_img_by_metal(channel_name)
        return imac

    def get_dict_stack_imgs(self, cond_list, channel_name, stack_name):
        plane_id = self.bro.helpers.dbhelp.get_plane_id(stack_name, channel_name)
        imac = {
            imid: np.flipud(self.stackimage.get_planeimg(int(imid), plane_id))
            for c, img_ids in cond_list
            for imid in img_ids
        }
        return imac

    @staticmethod
    def logvalue(val):
        new_val = np.log10(val + 0.1)

        return new_val

    def get_cond_id_im_id(self, condition_name):

        p = (
            self.session.query(
                db.images.image_id,
                db.conditions.condition_id,
            )
            .join(db.valid_images)
            .join(db.conditions)
            .filter(db.conditions.condition_name == condition_name)
        )

        pdat = self.bro.doquery(p)

        cond_id_im_id = []
        for cond, conddat in pdat.groupby("condition_id"):
            cond_im = (cond, conddat["image_id"].unique())
            cond_id_im_id.append(cond_im)

        return cond_id_im_id



import spherpro.bromodules.io_anndata as io_ann
import spherpro.bromodules.io_imcfolder as io_imc
import spherpro.bromodules.io_masks as io_masks
import spherpro.bromodules.io_stackimage as io_stackimage


class Io(object):
    def __init__(self, bro):
        self.masks = io_masks.IoMasks(bro)
        self.imcimg = io_imc.IoImc(bro)
        self.stackimg = io_stackimage.IoStackImage(bro)
        self.objmeasurements = io_ann.IoObjMeasurements(bro)



import plotnine as gg
import sqlalchemy as sa

import spherpro.bromodules.plot_base as plot_base
import spherpro.db as db

VAL_COL = db.object_measurements.value.key


class PlotScatter(plot_base.BasePlot):
    def __init__(self, bro):
        super().__init__(bro)
        # make the dependency explicit

        self.filter_measurements = self.bro.filters.measurements
        # define the measurement indexes with defaults
        # order is order expected by _get_measurement_filters
        self.measure_idx = [  # idx_name, default
            (db.objects.object_id.key, "cell"),
            (db.ref_planes.channel_name.key, None),
            (db.stacks.stack_name.key, "FullStack"),
            (db.measurement_names.measurement_name.key, "MeanIntensity"),
            (db.measurement_types.measurement_type.key, "Intensity"),
        ]

    def plot_bin2d(self, measure_x, measure_y, image_ids=None, filters=None):
        """
        Generates a plot where markers are plotted against each other in a 2D
        scatterplot
        Args:
            measure_x, measure_y: dict defining the selected measures:
                {db.objects.object_id.key: object_id,
                 db.ref_planes.channel_name.key: channel_name,
                 db.stacks.stack_name.key: stack_name,
                 db.measurement_names.measurement_name.key: measurement_name,
                 db.measurement_types.measurement_type.key: measurement_type}
        Returns:
            p:  the plot figure object

        """
        dat = self.get_marker_data(
            [measure_x, measure_y], image_ids=image_ids, filters=filters
        )
        p = (
            gg.ggplot(dat, gg.aes(x="0_" + VAL_COL, y="1_" + VAL_COL))
            + gg.geom_bin2d()
            + gg.geom_smooth(method="lm")
            + gg.xlab(" - ".join([measure_x.get(o, d) for o, d in self.measure_idx]))
            + gg.ylab(" - ".join([measure_y.get(o, d) for o, d in self.measure_idx]))
            + gg.scale_x_sqrt()
            + gg.scale_y_sqrt()
        )
        return p

    def plot_points(self, measure_x, measure_y, image_ids=None, filters=None):
        """
        Generates a plot where markers are plotted against each other in a 2D
        scatterplot
        Args:
            measure_x, measure_y: dict defining the selected measures:
                {db.objects.object_id.key: object_id,
                 db.ref_planes.channel_name.key: channel_name,
                 db.stacks.stack_name.key: stack_name,
                 db.measurement_names.measurement_name.key: measurement_name,
                 db.measurement_types.measurement_type.key: measurement_type}
            image_ids: image ids to plot
            filters: list of filter tuples [('filtername1', True),
            ('filtername2', False)]
        Returns:
            p:  the plot figure object

        """
        dat = self.get_marker_data(
            [measure_x, measure_y], image_ids=image_ids, filters=filters
        )
        p = (
            gg.ggplot(dat, gg.aes(x="0_" + VAL_COL, y="1_" + VAL_COL))
            + gg.geom_point()
            + gg.geom_smooth(method="lm")
            + gg.xlab(" - ".join([measure_x.get(o, d) for o, d in self.measure_idx]))
            + gg.ylab(" - ".join([measure_y.get(o, d) for o, d in self.measure_idx]))
            + gg.scale_x_sqrt()
            + gg.scale_y_sqrt()
        )
        return p

    def get_marker_data(self, measures, image_ids=None, filters=None):
        # TODO: this does not work with the current database structure!

        filters_measurement = [
            self.filter_measurements.to_delete1(
                *[[meas.get(o, d)] for o, d in self.measure_idx]
            )
            for meas in measures
        ]
        query = self._get_measurement_query()
        if image_ids is not None:
            query = query.filter(db.images.image_id.in_(image_ids))
        query_joins = self._get_joined_filtered_queries(query, filters_measurement)
        if filters is not None:
            # TODO: This needs to be fixed!
            query = query.filter(filters)
        dat = self.bro.doquery(query_joins)
        return dat

    def _get_joined_filtered_queries(self, base_query, filters, on_cols=None):
        """
        Queries repeatedly using the base_query and applying multiple filters.
        The results will be joined on the 'on_cols' and the columns renamed
        with a prefix according to the filter index (0_, 1_ etc.)

        Args:
            base_query: the query that should be used as a basis.
            filters:    a list of filter statements
            on_cols:    a list of column names. Default: ['ImageNumber',
            'ObjectNumber']
        Returns:
            the query to get the results
        """
        if on_cols is None:
            on_cols = [db.images.image_id.key, db.objects.object_number.key]
        queries = [
            base_query.filter(fil).subquery(name=str(i))
            for i, fil in enumerate(filters)
        ]
        # queries =[q.apply_labels() for q in queries]
        query_joins = self.session.query(*queries)
        for i in range(len(queries) - 1):
            query_joins = query_joins.join(
                queries[i + 1],
                (
                    sa.and_(
                        *[
                            getattr(queries[0].c, col) == getattr(queries[i + 1].c, col)
                            for col in on_cols
                        ]
                    )
                ),
            )
        query_joins = query_joins.with_labels()
        return query_joins

    def _get_measurement_data(self, filters):
        """
        Retrieves filtered measurement data
        """

        query = self._get_measurement_query()
        for fil in filters:
            query = query.filter(fil)
        dat = self.bro.doquery(query)
        return dat

    def _get_measurement_query(self):
        query = self.data.get_measurement_query().add_columns(
            db.images.image_id, db.objects.object_number
        )
        return query



import spherpro.bromodules.filter_objectfilters as filter_objectfilters

STACKNAME = "DistStack"
TYPENAME = "pixel_dist"
CHANNELNAME = "dist-rim"


class CalculateDistRim(object):
    """docstring for CalculateDistRim."""

    def __init__(self, bro):
        self.bro = bro
        self.session = self.bro.data.main_session
        self.data = self.bro.data
        self.custfilter = filter_objectfilters.ObjectFilterLib(bro)



import spherpro.bromodules.helpers_anndata as ha
import spherpro.bromodules.helpers_varia as hv


class Helpers(object):
    def __init__(self, bro):
        self.dbhelp = hv.HelperDb(bro)
        self.anndata = ha



import spherpro.bromodules.filter_measurements as filter_measurements
import spherpro.bromodules.filter_membership as filter_membership
import spherpro.bromodules.filter_objectfilters as filter_objectfilters


class Filters(object):
    def __init__(self, bro):
        self.membership = filter_membership.FilterMembership(bro)
        self.measurements = filter_measurements.FilterMeasurements(bro)
        self.objectfilterlib = filter_objectfilters.ObjectFilterLib(bro)



"""
A class to handle Anndata as a backend
"""
import pathlib
import shutil
import time

import anndata as ad
import numpy as np
import pandas as pd

import spherpro.bromodules.io_base as io_base
import spherpro.db as db

SUFFIX_ANNDATA = ".h5ad"


def get_anndata_filename(conf: object, object_type: str):
    fn = pathlib.Path(conf["sqlite"]["db"]).parent / (object_type + SUFFIX_ANNDATA)
    return fn


def scale_anndata(adat, col_scale=db.ref_stacks.scale.key, inplace=True):
    if not inplace:
        adat = adat.copy()
    adat.X = (adat.X.T * adat.var[col_scale][:, None]).T
    return adat


def copy_in_memory(adat):
    """
    Copies an adat into memory
    Args:
        adat: an andata object

    Returns:
        a in-memory copy of the anndata opbject

    """
    return ad.AnnData(np.array(adat.X), obs=adat.obs, var=adat.var, varm=adat.varm)


class IoAnnData(io_base.BaseIo):
    def __init__(self, bro, obj_type):
        super().__init__(bro)
        self.obj_type = obj_type
        self._adat = None

    @property
    def filename(self):
        """
        Filename of the file backing the anndata frame
        """
        return get_anndata_filename(self.bro.data.conf, self.obj_type)

    def initialize_anndata(self, data, var=None, obs=None):
        adat = ad.AnnData(data, var=var, obs=obs)
        adat.write(self.filename)

    @property
    def adat(self):
        a = self._adat
        if (a is None) or (  # Check if the data has been already loaded
            len(a.obs.index) != a.shape[0]
        ) | (
            len(a.var.index) != a.shape[1]
        ):  # check if data consistent
            self._adat = ad.read_h5ad(self.filename, backed="r")
        return self._adat


class IoObjMeasurements:
    def __init__(self, bro):
        self.bro = bro
        self._anndatadict = dict()
        self.scale_anndata = scale_anndata

    def get_anndata(self, obj_type):
        anndat = self._anndatadict.get(obj_type, None)
        if anndat is None:
            anndat = IoAnnData(self.bro, obj_type)
            self._anndatadict[obj_type] = anndat
        return anndat.adat

    def get_measurements(
        self,
        dat_obj=None,
        dat_meas=None,
        *,
        measidx=None,
        objidx=None,
        object_type=None,
        q_meas=None,
        q_obj=None,
    ):
        if q_meas is not None:
            dat_meas = self.bro.doquery(q_meas)
        if q_obj is not None:
            dat_obj = self.bro.doquery(q_obj)

        if dat_meas is not None:
            dat_meas = dat_meas.sort_values(db.measurements.measurement_id.key)
            measids = list(map(str, dat_meas[db.measurements.measurement_id.key]))
            dat_meas.index = measids
        else:
            measids = set(map(str, sorted(measidx)))

        if dat_obj is not None:
            obsidx = list(map(str, dat_obj[db.objects.object_id.key]))
            dat_obj.index = obsidx
            dat_obj.index.name = None
            dat_obj = dat_obj.sort_values(db.objects.object_id.key)
            it = dat_obj.groupby(db.objects.object_type.key)
        else:

            class obs:
                index = list(map(str, sorted(objidx)))

            it = [(object_type, obs)]

        dats = []
        for objtype, grp in it:
            obsidx = grp.index
            adat = self.get_anndata(objtype)
            varidx = list(set(measids).intersection(adat.var.index))
            if (len(varidx) > 0) & (len(obsidx) > 0):
                adat = adat[:, sorted(varidx, key=int)]
                dats.append(
                    ad.AnnData(adat.X, obs=adat.obs, var=adat.var)[obsidx, :].copy()
                )
        if len(dats) == 1:
            dat = dats[0]
        elif len(dats) == 0:
            raise ValueError("No valid measurements found")
        else:
            dat = ad.AnnData.concatenate(*dats, join="outer", index_unique=None)
        if dat_meas is not None:
            dat.var = dat.var.join(dat_meas)
        if dat_obj is not None:
            dat.obs = dat.obs.join(dat_obj)
        return dat

    @staticmethod
    def convert_anndata_legacy(adat):
        d = pd.DataFrame(
            adat.X, index=adat.obs.object_id, columns=adat.var.measurement_id
        ).stack()
        d.name = db.object_measurements.value.key
        dat = d.reset_index().merge(adat.obs.reset_index(drop=True)).merge(adat.var)
        return dat

    def add_anndata_datmeasurements(self, dat_new, replace=True, drop_all_old=False):
        for obj_type, dat in dat_new.groupby(db.objects.object_type.key):
            adat_new = ad.AnnData(
                dat.pivot(
                    index=db.objects.object_id.key,
                    columns=db.measurements.measurement_id.key,
                    values=db.object_measurements.value.key,
                )
            )
            self.add_anndata_objectmeasurements(obj_type, adat_new)

    def add_anndata_objectmeasurements(
        self, obj_type, adat_new, replace=True, drop_all_old=True
    ):
        """
        Adds a measurments to the anndata
        Args:
            obj_type: the object type
            adat_new: the anndata to be added
            replace: should existing measurements be updates?
            drop_all_old: should existing measurements be completly droped
                          before updating?

        Returns:

        """
        adat = self.get_anndata(obj_type)
        # Check that no new objects were added
        if len(get_difference(adat_new.obs.index, adat.obs.index)) != 0:
            raise ValueError(
                "The new data contains new objects, which is not supported yet."
            )

        old_vars = get_overlap(adat.var.index, adat_new.var.index)
        if len(old_vars) > 0:
            if not replace:
                raise ValueError(
                    f"Measurements {old_vars} already existing"
                    f" but replace=False was set!.\n"
                    f"Set replace=True to update values."
                )
            if (not drop_all_old) and (adat.shape[0] != adat_new.shape[0]):
                raise ValueError(
                    "Updating of existing variables only allowed"
                    " if values for all observations"
                    " are provided  or 'drop_all_old=True'"
                )
            else:
                kvars = [i for i in adat.var.index if i not in old_vars]
                adat = adat[:, kvars]

        adat = copy_in_memory(adat)
        adat = adat.T.concatenate(
            adat_new.T, index_unique=None, batch_key="batch", join="outer"
        ).T
        adat.var = adat.var.drop(columns="batch")

        ioan = self._anndatadict[obj_type]
        ioan.adat.file.close()

        fn_backup = f'{ioan.filename}.{time.strftime("%Y%m%d-%H%M%S")}'
        shutil.move(str(ioan.filename), fn_backup)

        # check order of variables
        ordvars = sorted(adat.var.index, key=int)
        if ordvars != list(adat.var.index):
            adat = adat[:, ordvars]

        adat.write(str(ioan.filename))
        ioan._adat = None
        return adat


def get_overlap(a, b):
    sa = set(a)
    sb = set(b)
    return sa.intersection(sb)


def get_difference(a, b):
    sa = set(a)
    sb = set(b)
    return sa.difference(sb)



import spherpro.bromodules.filters as filters
import spherpro.bromodules.helpers as helpers
import spherpro.bromodules.io as io
import spherpro.bromodules.plots as plots
import spherpro.bromodules.processing as processing
import spherpro.datastore as datastore
import spherpro.db as db


def get_bro(fn_config, readonly=True):
    """
    Convenience function to get a bro with a datastore initialized
    with a config file
    Args:
        fn_config: path to the config file
        readonly: load database readonly?
    Returns:
        A true bro
    """
    store = datastore.DataStore()
    store.read_config(fn_config)
    store.resume_data(readonly=readonly)
    bro = store.bro
    return bro


class Bro(object):
    """docstring for Bro."""

    def __init__(self, DataStore):
        self.data = DataStore
        self.helpers = helpers.Helpers(self)
        self.filters = filters.Filters(self)
        self.io = io.Io(self)
        self.plots = plots.Plots()
        self.plots.load_modules(self)
        self.processing = processing.Processing()
        self.processing.load_modules(self)
        self.doquery = self.data.query_df
        self.session = self.data.main_session

    @property
    def is_debarcoded(self):
        isdeb = False
        q = self.data.main_session.query(db.images.condition_id)
        q = q.filter(db.images.condition_id.isnot(None)).count()
        if q > 0:
            isdeb = True
        return isdeb



import sqlalchemy as sa

import spherpro.db as db


class HelperDb:
    def __init__(self, bro):
        self.bro = bro
        self.session = self.bro.data.main_session
        self.data = self.bro.data

    def get_target_by_channel(self, channel_name):
        target = (
            self.session.query(db.pannel.target)
            .filter(db.pannel.metal == channel_name)
            .first()
        )
        if target is None:
            target = [channel_name]
        return target[0]

    def get_objnumbers(self, obj_ids):
        bro = self.bro
        id_dict = {
            i: n
            for i, n in bro.session.query(
                db.objects.object_id, db.objects.object_number
            )
            .filter(db.objects.object_id.in_(obj_ids))
            .all()
        }
        return [id_dict[i] for i in obj_ids]

    def get_plane_id(self, stack_name: str, channel_name: str) -> int:
        """
        Get plane ID for given stack_name, channel_name
        Args:
            stack_name: A stack name
            channel_name: A channel name as given in the ref_planes table.

        Returns:
            The plane id

        """
        return (
            self.session.query(db.planes.plane_id)
            .join(db.ref_planes)
            .join(db.stacks)
            .filter(db.ref_planes.channel_name == channel_name)
            .filter(db.stacks.stack_name == stack_name)
        ).one()[0]

    def get_mask(self, img_id, object_type):
        bro = self.bro
        mask = bro.io.masks.get_mask(img_id, object_type)
        return mask

    def get_imc(self, img_id, channel):
        bro = self.bro
        return bro.io.imcimg.get_imcimg(img_id).get_img_by_metal(channel)

    def get_nb_dat(
        self, relationtype_name, obj_type=None, fil_query=None, valid_obj_only=True
    ):
        nbquery = (
            self.session.query(
                db.object_relations.object_id_parent,
                db.object_relations.object_id_child,
            )
            .join(db.object_relation_types)
            .filter(
                db.object_relation_types.object_relationtype_name == relationtype_name
            )
        )
        if valid_obj_only:
            c = sa.alias(db.valid_objects)
            p = sa.alias(db.valid_objects)
            nbquery = nbquery.filter(
                db.object_relations.object_id_child == c.c.object_id
            ).filter(db.object_relations.object_id_parent == p.c.object_id)
        if obj_type is not None:
            nbquery = nbquery.join(
                db.objects, db.objects.object_id == db.object_relations.object_id_parent
            ).filter(db.objects.object_type == obj_type)

        if fil_query is not None:
            q_fil = fil_query.alias()
            nbquery = nbquery.filter(
                db.object_relations.object_id_child == fil_query.c.object_id,
                db.object_relations.object_id_parent == q_fil.c.object_id,
            )
        return self.bro.doquery(nbquery)



# Different database collectors used by the datastore

from sqlalchemy import (
    Column,
    Integer,
    String,
    Float,
    Boolean,
    ForeignKeyConstraint,
    UniqueConstraint,
)
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
import sqlite3

Base = declarative_base()


# Define the table and column names to be used
# These need to match the definitions bellow


def connect_sqlite(conf):
    """
    creates a sqlite connector to be used with the Datastore.

    Args:

    Returns:
        SQLite3 conne:ctor
    """
    db = conf["sqlite"]["db"]
    conn = "sqlite:///%s" % (db)
    engine = create_engine(conn)
    Base.metadata.create_all(engine)
    return engine


def connect_sqlite_ro(conf):
    """
    creates a read only sqlite connector to be used with the Datastore.

    Args:

    Returns:
        SQLite3 conne:ctor
    """
    db = conf["sqlite"]["db"]

    def connect():
        return sqlite3.connect(f"file:{db}?mode=ro", uri=True)

    engine = create_engine("sqlite://", creator=connect)
    return engine


def connect_mysql(conf):
    """
    creates a MySQL connector to be used with the Datastore and creates the
    Database tables.

    Args:
        conf: the config dictionnary from a Datastore object.

    Returns:
        MySQL connector
    """
    host = conf["mysql"]["host"]
    port = conf["mysql"].get("port", "3306")
    user = conf["mysql"]["user"]
    password = conf["mysql"]["pass"]
    database = conf["mysql"]["db"]
    conn = "mysql+pymysql://%s:%s@%s:%s/%s" % (user, password, host, port, database)
    engine = create_engine(conn)
    return engine


def connect_postgresql(conf):
    """
    creates a MySQL connector to be used with the Datastore and creates the
    Database tables.

    Args:
        conf: the config dictionnary from a Datastore object.

    Returns:
        MySQL connector
    """
    conf_postgresql = "postgresql"
    host = conf[conf_postgresql]["host"]
    port = conf[conf_postgresql].get("port", "5432")
    user = conf[conf_postgresql]["user"]
    password = conf[conf_postgresql]["pass"]
    database = conf[conf_postgresql]["db"]
    conn = "postgresql+psycopg2://%s:%s@%s:%s/%s" % (
        user,
        password,
        host,
        port,
        database,
    )
    engine = create_engine(conn)
    return engine


def initialize_database(engine):
    Base.metadata.create_all(engine)
    return engine


def drop_all(conn):
    """
    drops all tables

    Args:
        conn: the connector.
    """
    Base.metadata.drop_all(conn)


################################################################################
#                           Model Definitions                                  #
################################################################################


class sampleblocks(Base):
    """
    Represents a physical sample block which is cut and placed onto
    slides.
    """

    __tablename__ = "sampleblocks"
    sampleblock_id = Column(Integer(), primary_key=True, autoincrement=True)
    sampleblock_name = Column(String(200))


class slides(Base):
    __tablename__ = "slides"
    slide_id = Column(Integer(), primary_key=True, autoincrement=True)
    slide_number = Column(String(200))
    sampleblock_id = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint([sampleblock_id], [sampleblocks.sampleblock_id]),
        {},
    )


class conditions(Base):
    """docstring for images."""

    __tablename__ = "conditions"
    condition_id = Column(Integer(), primary_key=True, autoincrement=True)
    condition_name = Column(String(200), server_default="default")
    time_point = Column(Float(), server_default="0.")
    barcode = Column(String(200))
    concentration = Column(Float(), server_default="0.")
    plate_id = Column(Integer(), server_default="1")
    bc_plate = Column(Integer(), server_default="0")
    bc_x = Column(Integer(), server_default="1")
    bc_y = Column(String(200), server_default="A")
    well_name = Column(String(200), server_default="A01")
    sampleblock_id = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint([sampleblock_id], [sampleblocks.sampleblock_id]),
        {},
    )


class slideacs(Base):
    __tablename__ = "slideacs"
    slideac_id = Column(Integer(), primary_key=True, autoincrement=True)
    slide_id = Column(Integer())
    slideac_name = Column(String(200))
    slideac_folder = Column(String(200))
    __table_args__ = (ForeignKeyConstraint([slide_id], [slides.slide_id]), {})


class sites(Base):
    """docstring for images."""

    __tablename__ = "sites"
    site_id = Column(Integer(), primary_key=True, autoincrement=True)
    slideac_id = Column(Integer())
    site_mcd_panoramaid = Column(Integer())
    site_name = Column(String(200))
    site_pos_x = Column(Integer())
    site_pos_y = Column(Integer())
    site_shape_h = Column(Integer())
    site_shape_w = Column(Integer())
    site_panorama = Column(String(200))
    __table_args__ = (ForeignKeyConstraint([slideac_id], [slideacs.slideac_id]), {})


class acquisitions(Base):
    __tablename__ = "acquisitions"
    acquisition_id = Column(Integer(), primary_key=True, autoincrement=True)
    site_id = Column(Integer())
    acquisition_mcd_acid = Column(Integer())
    acquisition_mcd_roiid = Column(Integer())
    acquisition_pos_x = Column(Integer())
    acquisition_pos_y = Column(Integer())
    acquisition_shape_h = Column(Integer())
    acquisition_shape_w = Column(Integer())
    acquisition_image_before = Column(String(200))
    acquisition_image_after = Column(String(200))
    acquisition_image_file = Column(String(200))
    __table_args__ = (ForeignKeyConstraint([site_id], [sites.site_id]), {})


class images(Base):
    """docstring for images."""

    __tablename__ = "images"
    image_id = Column(Integer(), primary_key=True, autoincrement=True)
    image_number = Column(Integer())
    image_pos_x = Column(Integer())
    image_pos_y = Column(Integer())
    image_shape_h = Column(Integer())
    image_shape_w = Column(Integer())
    crop_number = Column(Integer())
    acquisition_id = Column(Integer())
    bc_depth = Column(Float())
    bc_invalid = Column(Integer())
    bc_valid = Column(Integer())
    bc_highest_count = Column(Integer())
    bc_second_count = Column(Integer())
    condition_id = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint([condition_id], [conditions.condition_id]),
        ForeignKeyConstraint([acquisition_id], [acquisitions.acquisition_id]),
        {},
    )


TABLE_IMAGE = images.__tablename__


class masks(Base):
    """ a table describing the masks."""

    __tablename__ = "masks"
    object_type = Column(String(200), primary_key=True)
    image_id = Column(Integer(), primary_key=True)
    mask_filename = Column(String(200))
    __table_args__ = (ForeignKeyConstraint([image_id], [images.image_id]),)


class objects(Base):
    """docstring for objects."""

    __tablename__ = "objects"
    object_number = Column(Integer())
    object_id = Column(Integer(), primary_key=True, autoincrement=True)
    image_id = Column(Integer(), index=True)
    object_type = Column(String(200), index=True)
    __table_args__ = (
        ForeignKeyConstraint([image_id], [images.image_id]),
        ForeignKeyConstraint(
            [object_type, image_id], [masks.object_type, masks.image_id]
        ),
        {},
    )


class ref_stacks(Base):
    """docstring for ref_stacks."""

    __tablename__ = "ref_stacks"
    ref_stack_id = Column(Integer(), primary_key=True, autoincrement=True)
    ref_stack_name = Column(String(200), unique=True)
    scale = Column(Float())


class ref_planes(Base):
    """docstring for planes."""

    __tablename__ = "ref_planes"
    ref_stack_id = Column(Integer(), primary_key=True)
    ref_plane_number = Column(Integer(), primary_key=True)
    channel_type = Column(String(200))
    channel_name = Column(String(200))
    __table_args__ = (
        ForeignKeyConstraint([ref_stack_id], [ref_stacks.ref_stack_id]),
        {},
    )


class stacks(Base):
    """docstring for stacks."""

    __tablename__ = "stacks"
    stack_id = Column(Integer(), primary_key=True, autoincrement=True)
    stack_name = Column(String(200), unique=True)
    ref_stack_id = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint([ref_stack_id], [ref_stacks.ref_stack_id]),
        {},
    )


class planes(Base):
    __tablename__ = "planes"
    plane_id = Column(Integer(), primary_key=True, autoincrement=True)
    stack_id = Column(Integer())
    ref_plane_number = Column(Integer())
    ref_stack_id = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint(
            [ref_stack_id, ref_plane_number],
            [ref_planes.ref_stack_id, ref_planes.ref_plane_number],
        ),
        ForeignKeyConstraint([stack_id], [stacks.stack_id]),
        {},
    )


class modifications(Base):
    """docstring for modifications."""

    __tablename__ = "modifications"
    modification_id = Column(Integer(), primary_key=True, autoincrement=True)
    modification_name = Column(String(200), unique=True)
    modification_prefix = Column(String(200), unique=True)


class stack_modifications(Base):
    """docstring for stack_modifications."""

    __tablename__ = "stack_modifications"
    modification_id = Column(Integer(), primary_key=True)
    stack_id_parent = Column(Integer(), primary_key=True)
    stack_id_child = Column(Integer(), primary_key=True)
    __table_args__ = (
        ForeignKeyConstraint([stack_id_parent], [stacks.stack_id]),
        ForeignKeyConstraint([stack_id_child], [stacks.stack_id]),
        ForeignKeyConstraint([modification_id], [modifications.modification_id]),
        {},
    )


class image_stacks(Base):
    """
    Represents an image stack linked to an image
    """

    __tablename__ = "imagestacks"
    stack_id = Column(Integer(), primary_key=True)
    image_id = Column(Integer(), primary_key=True)
    image_stack_filename = Column(String(200))
    __table_args__ = (
        ForeignKeyConstraint([stack_id], [stacks.stack_id]),
        ForeignKeyConstraint([image_id], [images.image_id]),
        {},
    )


class object_filter_names(Base):
    __tablename__ = "object_filter_names"
    object_filter_id = Column(Integer(), primary_key=True, autoincrement=True)
    object_filter_name = Column(String(200), unique=True)


class object_filters(Base):
    __tablename__ = "object_filters"
    object_filter_id = Column(Integer(), primary_key=True)
    object_id = Column(Integer(), primary_key=True)
    filter_value = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint([object_id], [objects.object_id]),
        ForeignKeyConstraint(
            [object_filter_id], [object_filter_names.object_filter_id]
        ),
        {},
    )


class object_relation_types(Base):
    __tablename__ = "object_relation_types"
    object_relationtype_id = Column(Integer(), primary_key=True, autoincrement=True)
    object_relationtype_name = Column(String(200), index=True, unique=True)


class object_relations(Base):
    __tablename__ = "object_relations"
    object_id_parent = Column(Integer(), primary_key=True)
    object_id_child = Column(Integer(), primary_key=True)
    object_relationtype_id = Column(Integer(), primary_key=True)
    __table_args__ = (
        ForeignKeyConstraint([object_id_parent], [objects.object_id]),
        ForeignKeyConstraint([object_id_child], [objects.object_id]),
        ForeignKeyConstraint(
            [object_relationtype_id], [object_relation_types.object_relationtype_id]
        ),
        {},
    )


class measurement_names(Base):
    """
    Convenience table
    TODO: Consider removing
    """

    __tablename__ = "measurement_names"
    measurement_name = Column(String(200), primary_key=True)


class measurement_types(Base):
    """
    Convenience table
    TODO: Consider removing
    """

    __tablename__ = "measurement_types"
    measurement_type = Column(String(200), primary_key=True)


class measurements(Base):
    __tablename__ = "measurements"
    measurement_id = Column(Integer(), primary_key=True, autoincrement=True)
    measurement_type = Column(String(200))
    measurement_name = Column(String(200))
    plane_id = Column(Integer())
    __table_args__ = (
        ForeignKeyConstraint([measurement_name], [measurement_names.measurement_name]),
        ForeignKeyConstraint([measurement_type], [measurement_types.measurement_type]),
        ForeignKeyConstraint([plane_id], [planes.plane_id]),
        UniqueConstraint(measurement_name, measurement_type, plane_id),
        {},
    )


class object_measurements(Base):
    """docstring for object_measurements."""

    __tablename__ = "object_measurements"
    measurement_id = Column(Integer(), primary_key=True)
    object_id = Column(Integer(), primary_key=True)
    value = Column(Float(precision=32))
    __table_args__ = (
        ForeignKeyConstraint([measurement_id], [measurements.measurement_id]),
        ForeignKeyConstraint([object_id], [objects.object_id]),
        {},
    )


class pannel(Base):
    """docstring for pannel."""

    __tablename__ = "pannel"
    metal = Column(String(200), primary_key=True)
    target = Column(String(200), primary_key=True)
    antibody_clone = Column(String(200))
    concentration = Column(Float())
    is_ilastik = Column(Boolean())
    is_barcode = Column(Boolean())
    tube_number = Column(Integer())


class mask_measurements(Base):
    """docstring for image_measurements."""

    __tablename__ = "mask_measurements"
    image_id = Column(Integer(), primary_key=True)
    object_type = Column(String(200), primary_key=True)
    measurement_id = Column(Integer(), primary_key=True)
    value = Column(Float())
    __table_args__ = (
        ForeignKeyConstraint(
            [image_id, object_type], [masks.image_id, masks.object_type]
        ),
        ForeignKeyConstraint([measurement_id], [measurements.measurement_id]),
        {},
    )


class image_measurements(Base):
    """docstring for image_measurements."""

    __tablename__ = "image_measurements"
    image_id = Column(Integer(), primary_key=True)
    measurement_id = Column(Integer(), primary_key=True)
    value = Column(Float())
    __table_args__ = (
        ForeignKeyConstraint([image_id], [images.image_id]),
        ForeignKeyConstraint([measurement_id], [measurements.measurement_id]),
        {},
    )


class valid_images(Base):
    __tablename__ = "valid_images"
    image_id = Column(Integer(), primary_key=True)
    __table_args__ = (ForeignKeyConstraint([image_id], [images.image_id]), {})


class valid_objects(Base):
    __tablename__ = "valid_objects"
    object_id = Column(Integer(), primary_key=True)
    __table_args__ = (ForeignKeyConstraint([object_id], [objects.object_id]), {})



"""
A class to generate handle the loading of the stackimages specified in the database.
"""
import functools
import os

import numpy as np
import tifffile as tif

import spherpro.bromodules.io_base as io_base
import spherpro.configuration as conf
import spherpro.db as db

max_cache = 384

NCHANNEL_ERROR = "Image has incompatible number of channels."


class IoStackImage(io_base.BaseIo):
    def __init__(self, bro):
        super().__init__(bro)
        cpconf = self.data.conf[conf.CPOUTPUT]
        self.basedir = cpconf[conf.IMAGES_CSV][conf.STACKIMG_DIR]
        if self.basedir is None:
            self.basedir = self.data.conf[conf.CP_DIR]
        else:
            self.basedir = self.basedir.format(
                **{conf.CP_DIR: self.data.conf[conf.CP_DIR]}
            )
        self._dat_stackimgs = None

    @functools.lru_cache(maxsize=max_cache)
    def get_planeimg(self, image_id, plane_id):
        stack_id, plane_number = self._get_stackmeta_for_plane(plane_id)
        img = self.get_stackimg(image_id, stack_id)
        img_plane_number = plane_number - 1
        return img[img_plane_number, :, :]

    def get_stack_nchan(self, stack_id):
        nchan = (
            self.session.query(db.planes.plane_id)
            .join(db.stacks)
            .filter(db.stacks.stack_id == stack_id)
        ).count()
        return nchan

    def _get_stackmeta_for_plane(self, plane_id):
        stack_id, plane_number = (
            self.bro.session.query(
                db.planes.stack_id, db.planes.ref_plane_number
            ).filter(db.planes.plane_id == plane_id)
        ).one()
        return stack_id, plane_number

    @functools.lru_cache(maxsize=max_cache)
    def get_stackimg(self, image_id, stack_id=None):
        """
        Retrieves an image stack based on image & stack id
        Args:
            image_id: the image id
            stack_id: the stack id
        Keyword Args:
            stack_name: optionally accepted instead of stack_id
        Returns:
            image_array: a (memorymapped) image array
        """
        fn = self.get_stackimg_fn(image_id, stack_id)
        img = tif.imread(os.path.join(self.basedir, fn), out="memmap")
        imshape = img.shape
        nchan = self.get_stack_nchan(stack_id)

        if nchan == 1:
            img = img.squeeze()
            if len(img.shape) == 2:
                img = img.reshape([0] + list(img.shape))
            else:
                raise ValueError(NCHANNEL_ERROR)

        if imshape[0] == imshape[2]:
            if nchan < 4:
                last = True
            else:
                last = False
        elif imshape[2] == nchan:
            last = True
        elif imshape[0] == nchan:
            last = False
        elif nchan < 3:
            last = True

        else:
            raise ValueError(NCHANNEL_ERROR)
        if last:
            return np.rollaxis(img, 2)
        else:
            return img

    def get_stackimg_fn(self, image_id, stack_id=None, *, stack_name=None):
        q = self.bro.session.query(db.image_stacks.image_stack_filename).filter(
            db.image_stacks.image_id == image_id
        )
        if stack_id is not None:
            q = q.filter(db.image_stacks.stack_id == stack_id)
        elif stack_name is not None:
            q = q.join(db.stacks).filter(db.stacks.stack_name == stack_name)
        else:
            raise ValueError("Either stack_id or stack_name need to be provided.")
        fn = q.one()[0]
        return fn

    def clear_caches(self):
        self.get_stackimg.cache_clear()



import spherpro.bromodules.processing_debarcoding as processing_debarcoding
import spherpro.bromodules.processing_dist_rim as processing_dist_rim
import spherpro.bromodules.processing_measurementmaker as processing_mm
import spherpro.bromodules.processing_nb_agg as processing_nb_agg


class Processing(object):
    def __init__(self):
        pass

    def load_modules(self, bro):
        self.measurement_maker = processing_mm.MeasurementMaker(bro)
        self.debarcode = processing_debarcoding.Debarcode(bro)
        self.calculate_dist_rim = processing_dist_rim.CalculateDistRim(bro)
        self.nb_aggregation = processing_nb_agg.AggregateNeightbours(bro)



import operator

import numpy as np
import pandas as pd
import plotnine as gg

import spherpro as sp
import spherpro.configuration as conf
import spherpro.db as db

NAME_BARCODE = "BCString"
NAME_WELLCOLUMN = "CondID"
NAME_INVALID = "Invalid"
KEY_SECOND = db.images.bc_second_count.key
KEY_HIGHEST = db.images.bc_highest_count.key

SS_BARCODE_MEASUREMENT_NAME = "barcode"
SS_BARCODE_MEASUREMENT_TYPE = "object"


class Debarcode(object):
    """docstring for Debarcode."""

    def __init__(self, bro):
        self.bro = bro
        self.data = bro.data
        self.filter = sp.bromodules.filter_measurements.FilterMeasurements(self.bro)
        self.defaults_rawdist = bro.data.conf[conf.QUERY_DEFAULTS][conf.RAWDIST]
        self.defaults_channels = bro.data.conf[conf.QUERY_DEFAULTS][
            conf.CHANNEL_MEASUREMENTS
        ]
        self.DEFAULT_OBJTYPE = bro.data.conf[conf.QUERY_DEFAULTS][
            conf.DEFAULT_OBJECT_TYPE
        ]
        self.N = "n"
        self.COL_TYPE = "coltype"
        self.NOT_VALID = 0
        self.BC_METACOLS = [db.conditions.condition_id, db.conditions.sampleblock_id]
        self.COL_BC_METACOLS = [c.key for c in self.BC_METACOLS]
        self.CELL_METACOLS = [
            db.objects.object_id,
            db.images.image_id,
            db.sampleblocks.sampleblock_id,
        ]
        self.COL_CELL_METACOLS = [c.key for c in self.CELL_METACOLS]

    def debarcode(
        self,
        dist=None,
        borderdist=0,
        fils=None,
        stack=None,
        bc_treshs=None,
        measurement_name=None,
        transform=None,
    ):
        """
        Debarcodes the spheres in the dataset using the debarcoding information
        stored in the condition table
        """
        # get information from conditions and build the barcode key
        key = self._get_barcode_key()
        # get all intensities where dist-sphere<dist
        dat_cells = self._get_bc_cells(
            key,
            dist,
            fils=fils,
            borderdist=borderdist,
            stack=stack,
            measurement_name=measurement_name,
        )
        # threshold them
        dat_tresh = self._treshold_data(dat_cells, bc_treshs, transform)
        # debarcode them
        dat_db = self._debarcode_data(key, dat_tresh)
        # summarize them
        dat_sum = self._summarize_singlecell_barcodes(dat_db)
        # calculate statistics
        dat_stat = self._get_barcode_statistics(dat_sum)
        # update them to the Database
        self._write_bc(dat_stat, dist)
        # write single cell barcodes to the Database
        self._write_singlecell_barcodes(dat_db)

    def plot_histograms(
        self,
        dist=None,
        borderdist=0,
        fils=None,
        stack=None,
        measurement_name=None,
        transform="value",
    ):
        """
        Plot the histograms of the raw data
        """
        key = self._get_barcode_key()
        # get all intensities where dist-sphere<dist
        bcdat = self._get_bc_cells(
            key,
            dist,
            fils=fils,
            borderdist=borderdist,
            stack=stack,
            measurement_name=measurement_name,
            additional_meta=[db.sites.site_id],
        )
        bcvals = bcdat.stack()
        bcvals.name = "value"
        bcvals = bcvals.reset_index()
        bcvals["site"] = bcvals["site_id"].map(str)
        p = (
            gg.ggplot(bcvals, gg.aes(x=transform, color="site"))
            + gg.facet_wrap("channel_name", scales="free")
            + gg.geom_density()
        )
        return p

    def _write_bc(self, dat_bcstat, dist):
        """
        Writes the barcodes to the database
        """
        session = self.data.main_session

        # assert that debarcoding information from images not debarcoded
        # is deleted
        dat = dat_bcstat.set_index(db.images.image_id.key, drop=False)

        imgids = [i[0] for i in session.query(db.images.image_id).all()]
        dat = dat.reindex(index=imgids)
        for row in dat.to_dict(orient="records"):
            # convert all dtypes to int
            row.update(
                {str(c): int(v) if np.isfinite(v) else None for c, v in row.items()}
            )

            if row[db.images.condition_id.key] == self.NOT_VALID:
                # FIXTHIS: coding the invalid condition_id as 0 is dangerous!
                # However I do not know any better way to do this at the moment...
                row[db.images.condition_id.key] = None

            row[db.images.bc_depth.key] = dist
            session.query(db.images).filter(
                db.images.image_id == row[db.images.image_id.key]
            ).update(row)
        session.commit()

    @staticmethod
    def _default_treshfun(x):
        return x - np.mean(x) > 0

    def _treshold_data(
        self, bc_dat, bc_tresh=None, transform=None, meta_group=None, tresh_fun=None
    ):
        bc_dat = bc_dat.copy()
        if transform is not None:
            bc_dat = bc_dat.transform(transform)

        if tresh_fun is None:

            def tresh_fun(x):
                return (x - np.mean(x)) > 0

        if bc_tresh is None:
            if meta_group is not None:
                bc_dat = bc_dat.groupby(by=meta_group)
            bc_dat = bc_dat.apply(tresh_fun)
        else:
            t = np.array([bc_tresh[c] for c in bc_dat.columns])
            bc_dat = bc_dat.apply(lambda x: x > t, axis=1)
        bc_dat = bc_dat.astype(int)
        return bc_dat

    def _debarcode_data(self, bc_key, dat_tresh):
        meta_cols_bc = self.COL_BC_METACOLS
        meta_cols_cell = self.COL_CELL_METACOLS
        dat_db = (
            bc_key.reset_index(drop=False)
            .merge(
                dat_tresh.reset_index(drop=False),
                how="right",
            )
            .fillna(self.NOT_VALID)
            .loc[:, set(meta_cols_bc + meta_cols_cell)]
        )
        return dat_db

    def _summarize_singlecell_barcodes(self, dat_db):
        bc_meta_cols = self.COL_BC_METACOLS
        dat_sum = (
            dat_db.groupby(by=bc_meta_cols + [db.images.image_id.key])
            .size()
            .rename(self.N)
            .reset_index(drop=False)
        )
        return dat_sum

    def _get_barcode_statistics(self, dat_sum):
        dat_bcstat = (
            dat_sum.groupby(db.images.image_id.key)
            .apply(self._aggregate_barcodes)
            .reset_index(drop=True)
            .pivot_table(
                values=self.N,
                index=[db.images.image_id.key, db.conditions.condition_id.key],
                columns=self.COL_TYPE,
                fill_value=0,
            )
            .reset_index(drop=False)
        )
        return dat_bcstat

    def _aggregate_barcodes(self, dat):
        coltypelargest = [db.images.bc_highest_count.key, db.images.bc_second_count.key]
        fil = dat[db.conditions.condition_id.key] != self.NOT_VALID
        d_large = dat.loc[fil].nlargest(2, self.N).copy()
        d_large[self.COL_TYPE] = coltypelargest[: d_large.shape[0]]
        d_none = dat.loc[fil == False, :].copy()
        d_none[self.COL_TYPE] = db.images.bc_invalid.key
        d_sum = (
            dat.loc[fil]
            .groupby(db.images.image_id.key)[self.N]
            .sum()
            .rename(self.N)
            .reset_index(drop=False)
        )
        d_sum[self.COL_TYPE] = db.images.bc_valid.key
        d_out = pd.concat([d_large, d_none, d_sum], sort=True)
        if d_large.shape[0] > 0:
            cond = d_large[db.conditions.condition_id.key].iloc[0]
        else:
            cond = self.NOT_VALID
        d_out[db.conditions.condition_id.key] = cond
        return d_out

    def _get_barcode_key(self):
        bro = self.bro
        cond = bro.doquery(bro.session.query(*self.BC_METACOLS, db.conditions.barcode))
        cond = cond.set_index(self.COL_BC_METACOLS)
        key = cond[db.conditions.barcode.key].apply(lambda x: pd.Series(eval(x)))
        return key

    def _get_bc_cells(
        self,
        key,
        dist=None,
        fils=None,
        borderdist=0,
        stack=None,
        measurement_name=None,
        additional_meta=None,
    ):
        """
        Get cells for debarcoding
        Args:
            key: the barcoding key
            dist: the max distance to rim of the cells
            borderdist: the max distance, 0=only cells inside the sphere
            stack: the stack
            measurement_name: the measurement name
            additional_meta: a list of additional sqlalchemy columns to be added.
        returns:
            dat_bc: the single cell data
            dat_objmeta: the object metadata
            dat_meas: the measurment metadata
        """

        bro = self.bro
        if measurement_name is None:
            measurement_name = self.defaults_channels[conf.DEFAULT_MEASUREMENT_NAME]
        if stack is None:
            stack = self.defaults_channels[conf.DEFAULT_STACK_NAME]

        channels = tuple(key.columns.tolist())

        # Generate the measurement dict
        d_rawdist = self.defaults_rawdist
        filtdict = {
            db.stacks.stack_name.key: d_rawdist[conf.DEFAULT_STACK_NAME],
            db.ref_planes.channel_name.key: d_rawdist[conf.DEFAULT_CHANNEL_NAME],
            db.measurement_names.measurement_name.key: d_rawdist[
                conf.DEFAULT_MEASUREMENT_NAME
            ],
        }
        dist_measid = self.filter.measmeta_to_measid(**filtdict)

        q_obj = (
            self.data.get_objectmeta_query()
            .filter(db.objects.object_type == self.DEFAULT_OBJTYPE)
            .add_columns(db.sampleblocks.sampleblock_id)
        )

        if additional_meta is not None:
            q_obj = q_obj.add_columns(*additional_meta)
        dat_obj = bro.doquery(q_obj)

        dat_filmeas = bro.doquery(
            self.data.get_measmeta_query()
            .filter(db.measurements.measurement_id == dist_measid)
            .add_columns(db.ref_stacks.scale)
        )
        dat_fil = bro.io.objmeasurements.get_measurements(dat_obj, dat_filmeas)
        dat_fil = bro.io.objmeasurements.scale_anndata(dat_fil)

        # Get the distance filters
        distfils = [(dist_measid, operator.gt, borderdist)]
        if dist is not None:
            distfils += [(dist_measid, operator.lt, dist)]

        dat_obj = dat_fil.obs.loc[
            bro.filters.measurements.get_filter_vector(dat_fil, distfils), :
        ]
        # get the data query
        fil_meas = bro.filters.measurements.get_measmeta_filter_statements(
            channel_names=[channels],
            stack_names=[stack],
            measurement_names=[measurement_name],
            measurement_types=[None],
        )
        dat_meas = bro.doquery(
            self.data.get_measmeta_query()
            .filter(fil_meas)
            .add_columns(db.ref_stacks.scale, db.ref_planes.channel_name)
        )
        dat_cells = bro.io.objmeasurements.get_measurements(dat_obj, dat_meas)
        dat_cells = bro.io.objmeasurements.scale_anndata(dat_cells)

        dat_bccells = pd.DataFrame(
            dat_cells.X,
            index=pd.MultiIndex.from_frame(dat_cells.obs),
            columns=dat_cells.var[db.ref_planes.channel_name.key],
        )
        return dat_bccells

    def _write_singlecell_barcodes(self, dat_db):
        """
        Saves the singlecell barcodes to the database
        """
        mm = self.bro.processing.measurement_maker
        dat_db = dat_db.copy()
        val_obj = self.bro.data.conf[conf.QUERY_DEFAULTS][conf.OBJECT_DEFAULTS]

        plane_id = mm.get_object_plane_id()
        meas_id = mm.register_single_measurement(
            SS_BARCODE_MEASUREMENT_NAME, SS_BARCODE_MEASUREMENT_TYPE, plane_id
        )
        dat_db[db.measurements.measurement_id.key] = meas_id
        dat_db = dat_db.rename(
            columns={db.conditions.condition_id.key: db.object_measurements.value.key}
        )
        dat_db[db.objects.object_type.key] = self.DEFAULT_OBJTYPE
        self.bro.processing.measurement_maker.add_object_measurements(
            dat_db, drop_all_old=True
        )



import spherpro.db as db


class HelpersAnndata:
    def __init__(self, bro):
        self.bro = bro
        self.session = self.bro.data.main_session
        self.data = self.bro.data


def add_anndata_varmeta(ad, dat_meta, on=None, on_index=False):
    ad.var = _add_anndata_meta(
        ad.var,
        dat_meta,
        index_var=db.measurements.measurement_id.key,
        on=on,
        on_index=on_index,
    )
    return ad


def add_anndata_obsmeta(ad, dat_meta, on=None, on_index=False):
    ad.obs = _add_anndata_meta(
        ad.obs, dat_meta, index_var=db.objects.object_id.key, on=on, on_index=on_index
    )
    return ad


def _add_anndata_meta(ad_meta, dat_meta, index_var, on=None, on_index=False):
    if (
        (on_index)
        or ((on is None) and index_var in dat_meta.columns)
        or (on == index_var)
    ):
        ad_meta = _join_indexvar(ad_meta, dat_meta, index_var, on_index)
    else:
        ad_meta = _merge_col(ad_meta, dat_meta, on=on)
    return ad_meta


def _merge_col(admeta, dat, on=None):
    admeta.index.name = "index"
    return (admeta.reset_index().merge(dat, on=on).set_index("index")).loc[
        admeta.index, :
    ]


def _join_indexvar(admeta, dat, idxvar, on_index):
    if on_index is False:
        dat = dat.set_index(idxvar)
    dat.index = map(str, dat.index)
    return admeta.join(dat, how="left")



import spherpro.bromodules.plot_debarcodequality as plot_debarcodequality
import spherpro.bromodules.plot_heatmask as plot_heatmask
import spherpro.bromodules.plot_scatterplot as plot_scatterplot


class Plots(object):
    def __init__(self):
        pass

    def load_modules(self, bro):
        self.scatterplot = plot_scatterplot.PlotScatter(bro)
        self.heatmask = plot_heatmask.PlotHeatmask(bro)
        self.debarcoedequality = plot_debarcodequality.PlotDebarcodeQuality(bro)
        self.debarcoededcells = plot_debarcodequality.PlotDebarcodeCells(bro)



import logging
import os
import re
import warnings

# import numpy as np
from os import listdir
from os.path import isfile, join

import numpy as np
import pandas as pd
import sqlalchemy as sa

# from odo import odo
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import sessionmaker

import spherpro.bro as bro
import spherpro.bromodules.io_anndata as io_anndata
import spherpro.configuration as config
import spherpro.db as db
import spherpro.library as lib

DICT_DB_KEYS = {
    "image_id": db.images.image_id.key,
    "object_number": db.objects.object_number.key,
    "measurement_type": db.measurement_types.measurement_type.key,
    "measurement_name": db.measurement_names.measurement_name.key,
    "stack_name": db.stacks.stack_name.key,
    "plane_id": db.ref_planes.ref_plane_number.key,
    "object_id": db.objects.object_id.key,
}

OBJECTS_STACKNAME = "ObjectStack"
OBJECTS_CHANNELNAME = "object"
OBJECTS_PLANEID = "1"
OBJECTS_CHANNELTYPE = "object"

READONLY = "_readonly"


class DataStore(object):
    """DataStore
    The DataStore class is intended to be used as a storage for spheroid IMC
    data.

    Methods:
        Base:
            read_config: read configfile
            import_data: reads and writes data to the database
            resume_data: reads non-database files and configures backend
    """

    def __init__(self):
        # init empty properties here
        self.experiment_layout = None
        self.barcode_key = None
        self.well_measurements = None
        self.cut_meta = None
        self.roi_meta = None
        self.channel_meta = None
        self.sphere_meta = None
        self.measurement_meta_cache = None
        self._pannel = None
        self._session = None
        self._session_maker = None
        self.connectors = {
            config.CON_SQLITE: db.connect_sqlite,
            config.CON_SQLITE + READONLY: db.connect_sqlite_ro,
            config.CON_MYSQL: db.connect_mysql,
            config.CON_POSTGRESQL: db.connect_postgresql,
        }

    #########################################################################
    #########################################################################
    #                      Import or Resume functions:                      #
    #########################################################################
    #########################################################################

    def read_config(self, configpath):
        """
        finds the measurement meta information from a given string

        Args:
            configpath: A string denoting the location of the config file

        Raises:
            YAMLError
        """
        self.conf = config.read_configuration(configpath)

    def import_data(self, minimal=None):
        """read_data
        Reads the Data using the file locations given in the configfile.
        Args:
            minimal: Bool, if True, the import process only imports values from
                the RefStacks and no location values
        """
        if minimal is None:
            minimal = False
        # Read the data based on the config
        self._read_experiment_layout()
        self._read_barcode_key()
        # self._read_measurement_data()
        self._read_image_data()
        self._read_relation_data()
        self._read_stack_meta()
        self._populate_db(minimal)

    def resume_data(self, readonly=False):
        """read_data
        Reads non-database files and configures backend according to
        the configfile.
        """
        # Read the data based on the config
        # self._read_experiment_layout()
        # self._read_barcode_key()
        # self._readWellMeasurements()
        # self._read_cut_meta()
        # self._read_roi_meta()
        # self._read_measurement_data()
        # self._read_stack_meta()
        self._read_pannel()
        backend = self.conf[config.BACKEND]
        if readonly:
            backend += READONLY

        self.db_conn = self.connectors[backend](self.conf)
        self.bro = bro.Bro(self)

    def drop_all(self):
        self.db_conn = self.connectors[self.conf[config.BACKEND]](self.conf)
        db.drop_all(self.db_conn)

    ##########################################
    #   Helper functions used by readData:   #
    ##########################################

    def _read_experiment_layout(self):
        """
        reads the experiment layout as stated in the config
        and saves it in the datastore
        """
        if self.conf[config.LAYOUT_CSV][config.PATH] is not None:
            sep = self.conf[config.LAYOUT_CSV][config.SEP]
            experiment_layout = pd.read_csv(
                self.conf[config.LAYOUT_CSV][config.PATH], sep=sep
            )
            # rename the columns
            rename_dict = {
                self.conf[config.LAYOUT_CSV][c]: target
                for c, target in [
                    (config.LAYOUT_CSV_COND_ID, db.conditions.condition_id.key),
                    (config.LAYOUT_CSV_COND_NAME, db.conditions.condition_name.key),
                    (config.LAYOUT_CSV_TIMEPOINT_NAME, db.conditions.time_point.key),
                    (config.LAYOUT_CSV_BARCODE, db.conditions.barcode.key),
                    (
                        config.LAYOUT_CSV_CONCENTRATION_NAME,
                        db.conditions.concentration.key,
                    ),
                    (config.LAYOUT_CSV_BC_PLATE_NAME, db.conditions.bc_plate.key),
                    (config.LAYOUT_CSV_PLATE_NAME, db.conditions.plate_id.key),
                    (config.LAYOUT_CSV_WELL_NAME, db.conditions.well_name.key),
                ]
            }
            experiment_layout = experiment_layout.rename(columns=rename_dict)
            self.experiment_layout = experiment_layout.fillna(0)
        else:
            self.experiment_layout = None

    def _read_barcode_key(self):
        """
        reads the barcode key as stated in the config
        """
        conf_bc = self.conf[config.BARCODE_CSV]
        conf_layout = self.conf[config.LAYOUT_CSV]
        path = conf_bc[config.PATH]
        if path is not None:
            # Load the barcode key
            sep = conf_bc[config.SEP]
            barcodes = pd.read_csv(path, sep=sep)
            # Adapt the names
            rename_dict = {
                conf_bc[config.BC_CSV_PLATE_NAME]: conf_layout[
                    config.LAYOUT_CSV_BC_PLATE_NAME
                ],
                conf_bc[config.BC_CSV_WELL_NAME]: db.conditions.well_name.key,
            }
            barcodes = barcodes.rename(columns=rename_dict)
            # Convert the barcode key to a dictionary string
            barcodes = barcodes.set_index(list(rename_dict.values()))
            barcodes = (
                barcodes.transpose()
                # converts the barcodes to a string dictionary
                .apply(lambda x: str(x.to_dict()))
            )
            barcodes = barcodes.rename(db.conditions.barcode.key)
            barcodes = barcodes.reset_index(drop=False)
            self.barcode_key = barcodes
        else:
            self.barcode_key = None

    def _read_objtype_measurements(self, object_type, chunksize):
        conf_meas = self.conf[config.CPOUTPUT][config.MEASUREMENT_CSV]
        sep = conf_meas[config.SEP]
        cpdir = self.conf[config.CP_DIR]
        filetype = conf_meas[config.FILETYPE]
        reader = pd.read_csv(
            os.path.join(cpdir, object_type + filetype), sep=sep, chunksize=chunksize
        )

        if chunksize is None:
            reader = [reader]
        for dat_objmeas in reader:
            rename_dict = {
                self.conf[config.OBJECTNUMBER]: db.objects.object_number.key,
                self.conf[config.IMAGENUMBER]: db.images.image_number.key,
            }
            dat_objmeas.rename(columns=rename_dict, inplace=True)
            dat_objmeas[db.objects.object_type.key] = object_type
            yield dat_objmeas

    def _read_image_data(self):
        cpdir = self.conf[config.CP_DIR]
        rename_dict = {self.conf[config.IMAGENUMBER]: db.images.image_number.key}
        images_csv = lib.read_csv_from_config(
            self.conf[config.CPOUTPUT][config.IMAGES_CSV], base_dir=cpdir
        )
        images_csv = images_csv.rename(columns=rename_dict)
        self._images_csv = images_csv

    def _read_relation_data(self):
        conf_rel = self.conf[config.CPOUTPUT][config.RELATION_CSV]
        cpdir = self.conf[config.CP_DIR]
        relation_csv = lib.read_csv_from_config(
            self.conf[config.CPOUTPUT][config.RELATION_CSV], base_dir=cpdir
        )
        col_map = {
            conf_rel[c]: target
            for c, target in [
                (config.OBJECTTYPE_FROM, config.OBJECTTYPE_FROM),
                (config.OBJECTTYPE_TO, config.OBJECTTYPE_TO),
                (config.OBJECTNUMBER_FROM, config.OBJECTNUMBER_FROM),
                (config.OBJECTNUMBER_TO, config.OBJECTNUMBER_TO),
                (config.IMAGENUMBER_FROM, config.IMAGENUMBER_FROM),
                (config.IMAGENUMBER_TO, config.IMAGENUMBER_TO),
                (
                    config.RELATIONSHIP,
                    db.object_relation_types.object_relationtype_name.key,
                ),
            ]
        }

        self._relation_csv = relation_csv.rename(columns=col_map)

    def _read_stack_meta(self):
        """
        reads the stack meta as stated in the config
        and saves it in the datastore
        """
        stack_dir = self.conf[config.STACK_DIR][config.PATH]
        sep = self.conf[config.STACK_DIR][config.SEP]
        match = re.compile("(.*)\.csv")
        stack_files = [f for f in listdir(stack_dir) if isfile(join(stack_dir, f))]
        stack_data = [pd.read_csv(join(stack_dir, n), sep) for n in stack_files]
        stack_files = [match.match(name).groups()[0] for name in stack_files]
        self.stack_csvs = {stack: data for stack, data in zip(stack_files, stack_data)}
        self._stack_relation_csv = lib.read_csv_from_config(
            self.conf[config.STACK_RELATIONS]
        )

    def _read_pannel(self):
        """
        Reads the pannel as stated in the config.
        """
        self._pannel = lib.read_csv_from_config(self.conf[config.PANNEL_CSV])

    def _populate_db(self, minimal):
        """
        Writes the tables to the database
        """
        self.db_conn = self.connectors[self.conf[config.BACKEND]](self.conf)
        self.drop_all()
        db.initialize_database(self.db_conn)

        self.bro = bro.Bro(self)
        self._write_imagemeta_tables()
        self._write_masks_table()
        self._write_stack_tables()
        self._write_refplanes_table()
        self._write_planes_table()
        self._write_pannel_table()
        self._write_condition_table()
        self._write_measurement_table(minimal)
        self._write_image_stacks_table()
        self.reset_valid_objects()
        self.reset_valid_images()
        # self._write_object_relations_table()

    #### Helpers ####

    def replace_condition_table(self):
        """
        This is used in case an the experiment layout or
        barcoding is updated.

        Note that this will delete any debarcoding.
        """
        # read the tables
        self._read_experiment_layout()
        self._read_barcode_key()

        # delete the link between images and conditions
        session = self.main_session
        q = session.query(db.images).update({db.images.condition_id.key: None})
        # delete the existing table
        (session.query(db.conditions).delete())
        session.commit()

        # write the table
        self._write_condition_table()
        session.commit()

    ##########################################
    #        Database Table Generation:      #
    ##########################################

    def _write_stack_tables(self):
        """
        Creates the StackModifications, StackRelations, Modifications,
        RefStack and DerivedStack tables and writes them to the database
        """

        # Modifications
        modifications = self._generate_modifications()
        self._bulkinsert(modifications, db.modifications)

        # RefStacks
        refstack = self._generate_refstack()
        self._bulkinsert(refstack, db.ref_stacks)

        # Stacks
        stack = self._generate_stack()
        self._bulkinsert(stack, db.stacks)

        # StackModifications
        stackmodification = self._generate_stackmodification()
        self._bulkinsert(stackmodification, db.stack_modifications)

    def _generate_modifications(self):
        """
        Generates the modification table
        """
        parent_col = self.conf[config.STACK_RELATIONS][config.PARENT]
        modname_col = self.conf[config.STACK_RELATIONS][config.MODNAME]
        modpre_col = self.conf[config.STACK_RELATIONS][config.MODPRE]
        stackrel = self._stack_relation_csv.loc[
            self._stack_relation_csv[parent_col] != "0"
        ]
        Modifications = pd.DataFrame(stackrel[modname_col])
        Modifications["tmp"] = stackrel[modpre_col]
        Modifications.columns = [
            db.modifications.modification_name.key,
            db.modifications.modification_prefix.key,
        ]
        Modifications[db.modifications.modification_id.key] = self._query_new_ids(
            db.modifications.modification_id, Modifications.shape[0]
        )
        return Modifications

    def _generate_stackmodification(self):
        """
        generates the stackmodification table
        """
        parent_col = self.conf[config.STACK_RELATIONS][config.PARENT]
        modname_col = self.conf[config.STACK_RELATIONS][config.MODNAME]
        stack_col = self.conf[config.STACK_RELATIONS][config.STACK]
        key_map = {
            parent_col: db.stack_modifications.stack_id_parent.key,
            modname_col: db.modifications.modification_name.key,
            stack_col: db.stack_modifications.stack_id_child.key,
        }

        StackModification = self._stack_relation_csv.loc[
            self._stack_relation_csv[parent_col] != "0", list(key_map.keys())
        ].rename(columns=key_map)

        stackdict = {
            n: i
            for n, i in self.main_session.query(
                db.stacks.stack_name, db.stacks.stack_id
            ).filter(
                db.stacks.stack_name.in_(
                    StackModification[
                        db.stack_modifications.stack_id_parent.key
                    ].tolist()
                    + StackModification[
                        db.stack_modifications.stack_id_child.key
                    ].tolist()
                )
            )
        }
        StackModification[
            db.stack_modifications.stack_id_parent.key
        ] = StackModification[db.stack_modifications.stack_id_parent.key].replace(
            stackdict
        )

        StackModification[
            db.stack_modifications.stack_id_child.key
        ] = StackModification[db.stack_modifications.stack_id_child.key].replace(
            stackdict
        )
        modidict = {
            n: i
            for n, i in (
                self.main_session.query(
                    db.modifications.modification_name, db.modifications.modification_id
                ).filter(
                    db.modifications.modification_name.in_(
                        StackModification[db.modifications.modification_name.key]
                    )
                )
            )
        }

        StackModification[db.modifications.modification_id.key] = StackModification[
            db.modifications.modification_name.key
        ].replace(modidict)

        return StackModification.loc[
            :,
            [
                db.stack_modifications.stack_id_parent.key,
                db.stack_modifications.stack_id_child.key,
                db.modifications.modification_id.key,
            ],
        ]

    def _generate_refstack(self):
        """
        Generates the refstack table
        """
        stack_col = self.conf[config.STACK_RELATIONS][config.STACK]
        ref_col = self.conf[config.STACK_RELATIONS][config.REF]
        key_map = {stack_col: db.ref_stacks.ref_stack_name.key}

        ref_stack = self._stack_relation_csv.loc[
            self._stack_relation_csv[ref_col] == "0", list(key_map.keys())
        ].rename(columns=key_map)
        scale_col = self.conf[config.CPOUTPUT][config.IMAGES_CSV][config.SCALING_PREFIX]
        scale_names = [
            scale_col + n for n in ref_stack[db.ref_stacks.ref_stack_name.key]
        ]
        dat_img = self._images_csv.reindex(columns=scale_names, fill_value=1)
        scales = dat_img.iloc[0, :]
        # assert that scales are equal in all images
        assert np.all(dat_img.eq(scales, axis=1))
        ref_stack[db.ref_stacks.scale.key] = scales.values
        ref_stack = ref_stack.append(
            pd.DataFrame(
                {
                    db.ref_stacks.ref_stack_name.key: OBJECTS_STACKNAME,
                    db.ref_stacks.scale.key: 1,
                },
                index=[1],
            ),
            ignore_index=True,
        )
        # set uni id
        ref_stack[db.ref_stacks.ref_stack_id.key] = self._query_new_ids(
            db.ref_stacks.ref_stack_id, (ref_stack.shape[0])
        )
        return ref_stack

    def _generate_stack(self):
        """
        Genes the DerivedStack
        """
        stack_col = self.conf[config.STACK_RELATIONS][config.STACK]
        ref_col = self.conf[config.STACK_RELATIONS][config.REF]
        key_map = {
            stack_col: db.stacks.stack_name.key,
            ref_col: db.ref_stacks.ref_stack_name.key,
        }

        stack = self._stack_relation_csv.loc[:, list(key_map.keys())].rename(
            columns=key_map
        )
        # Add the 'objects' stack
        stack = stack.append(
            {
                db.stacks.stack_name.key: OBJECTS_STACKNAME,
                db.ref_stacks.ref_stack_name.key: OBJECTS_STACKNAME,
            },
            ignore_index=True,
        )

        fil = stack[db.ref_stacks.ref_stack_name.key] == "0"
        stack.loc[fil, db.ref_stacks.ref_stack_name.key] = stack.loc[
            fil, db.stacks.stack_name.key
        ]

        refstackdict = {
            n: i
            for n, i in (
                self.main_session.query(
                    db.ref_stacks.ref_stack_name, db.ref_stacks.ref_stack_id
                ).filter(
                    db.ref_stacks.ref_stack_name.in_(
                        stack[db.ref_stacks.ref_stack_name.key]
                    )
                )
            )
        }
        stack[db.ref_stacks.ref_stack_id.key] = stack[
            db.ref_stacks.ref_stack_name.key
        ].replace(refstackdict)

        stack[db.stacks.stack_id.key] = self._query_new_ids(
            db.stacks.stack_id, (stack.shape[0])
        )

        return stack

    def _write_refplanes_table(self):
        """
        generates the PlaneMeta Table and writes it to the database.
        """
        planes = self._generate_refplanemeta()
        self._bulkinsert(planes, db.ref_planes)

    def _write_planes_table(self):
        """
        generates the PlaneMeta Table and writes it to the database.
        """
        planes = self._generate_planemeta()
        self._bulkinsert(planes, db.planes)

    def _generate_refplanemeta(self):

        stack_col = self.conf[config.STACK_DIR][config.STACK]
        id_col = self.conf[config.STACK_DIR][config.ID]
        name_col = self.conf[config.STACK_DIR][config.NAME]
        type_col = self.conf[config.STACK_DIR][config.TYPE]
        planes = pd.DataFrame(
            columns=[
                db.ref_planes.ref_plane_number.key,
                db.ref_stacks.ref_stack_name.key,
                db.ref_planes.channel_name.key,
                db.ref_planes.channel_type.key,
            ]
        )
        for stack in self.stack_csvs:
            self.stack_csvs[stack].rename(
                columns={
                    id_col: db.ref_planes.ref_plane_number.key,
                    stack_col: db.ref_stacks.ref_stack_name.key,
                    name_col: db.ref_planes.channel_name.key,
                    type_col: db.ref_planes.channel_type.key,
                },
                inplace=True,
            )
            planes = planes.append(self.stack_csvs[stack])
        planes = planes.reset_index()
        del planes["index"]
        # cast PlaneID to be identical to the one in Measurement:
        planes[db.ref_planes.ref_plane_number.key] = planes[
            db.ref_planes.ref_plane_number.key
        ].apply(lambda x: int(x))

        planes = planes.append(
            {
                db.ref_planes.ref_plane_number.key: OBJECTS_PLANEID,
                db.ref_stacks.ref_stack_name.key: OBJECTS_STACKNAME,
                db.ref_planes.channel_name.key: OBJECTS_CHANNELNAME,
                db.ref_planes.channel_type.key: OBJECTS_CHANNELTYPE,
            },
            ignore_index=True,
        )
        refdict = self._get_namekey_dict(
            db.ref_stacks.ref_stack_name,
            db.ref_stacks.ref_stack_id,
            planes[db.ref_stacks.ref_stack_name.key].unique(),
        )

        planes[db.ref_stacks.ref_stack_id.key] = planes[
            db.ref_stacks.ref_stack_name.key
        ].replace(refdict)

        return planes.loc[
            :,
            [
                db.ref_stacks.ref_stack_id.key,
                db.ref_planes.ref_plane_number.key,
                db.ref_planes.channel_name.key,
                db.ref_planes.channel_type.key,
            ],
        ]

    def _generate_planemeta(self):
        stack = self._generate_stack()
        refplanes = self._generate_refplanemeta()
        planes = stack.merge(refplanes, on=db.ref_stacks.ref_stack_id.key)
        stackdic = self._get_namekey_dict(
            db.stacks.stack_name,
            db.stacks.stack_id,
            planes[db.stacks.stack_name.key].unique().tolist(),
        )
        planes[db.stacks.stack_id.key] = planes[db.stacks.stack_name.key].replace(
            stackdic
        )
        planes = planes.loc[
            :,
            [
                db.stacks.stack_id.key,
                db.ref_stacks.ref_stack_id.key,
                db.ref_planes.ref_plane_number.key,
            ],
        ]
        planes[db.planes.plane_id.key] = self._query_new_ids(
            db.planes.plane_id, (planes.shape[0])
        )
        return planes

    def _write_imagemeta_tables(self):
        """
        Write the tables containing the image metadata
        This contains:
            - the acquisition ROI 'roi': original acquisition where the image
                was cropped from
            - the site: the site on the slide where the acquisition ROI was made
                -> corresponds to a panroma in the MCD
            - the slide: the physical slide
        """
        dat_image = self._generate_image_table()
        dat_image, dat_roi = self._generate_roi_table(dat_image)
        dat_roi, dat_site = self._generate_site_table(dat_roi)
        dat_site, dat_slideac = self._generate_slideac_table(dat_site)
        dat_slideac, dat_slide = self._generate_slide_table(dat_slideac)
        dat_slide, dat_sampleblock = self._generate_sampleblock_table(dat_slide)

        self._bulkinsert(dat_sampleblock, db.sampleblocks)
        self._bulkinsert(dat_slide, db.slides)
        self._bulkinsert(dat_slideac, db.slideacs)
        self._bulkinsert(dat_site, db.sites)
        self._bulkinsert(dat_roi, db.acquisitions)
        self._bulkinsert(dat_image, db.images)
        self._bulkinsert(dat_image, db.valid_images)

    def _generate_image_table(self):
        """
        Generates the slide, site and roi metadata from the filenames or ome folders.
        """
        cpconf = self.conf[config.CPOUTPUT]
        imgconf = cpconf[config.IMAGES_CSV]
        objects = cpconf[config.MEASUREMENT_CSV][config.OBJECTS]
        obj = objects[0]
        prefix = cpconf[config.IMAGES_CSV][config.MASK_FILENAME_PREFIX]
        # use any object to get a filename
        rename_dict = {
            prefix + obj: "fn",
            imgconf[config.IMAGE_HEIGHT_PREFIX] + obj: db.images.image_shape_h.key,
            imgconf[config.IMAGE_WIDTH_PREFIX] + obj: db.images.image_shape_w.key,
        }
        dat_fn = self._images_csv.loc[
            :, [db.images.image_number.key] + list(rename_dict.keys())
        ]
        dat_fn = dat_fn.rename(columns=rename_dict)

        re_meta = imgconf[config.META_REGEXP]
        img_meta = lib.map_group_re(dat_fn["fn"], re_meta)
        img_meta.index = dat_fn.index
        img_meta = img_meta.join(dat_fn)

        colmap = {
            imgconf[g]: v
            for g, v in [
                (config.GROUP_POSX, db.images.image_pos_x.key),
                (config.GROUP_POSY, db.images.image_pos_y.key),
                (config.GROUP_CROPID, db.images.crop_number.key),
                (config.GROUP_BASENAME, config.GROUP_BASENAME),
            ]
        }
        img_meta = img_meta.rename(columns=colmap)
        img_meta[db.images.image_id.key] = self._query_new_ids(
            db.images.image_id, img_meta.shape[0]
        )
        return img_meta

    def _generate_roi_table(self, img_meta):
        """
        Generates the ROi metadata table and updates the img_meta table to link
        to the roi_table
        """
        imgconf = self.conf[config.CPOUTPUT][config.IMAGES_CSV]
        roi_basenames = img_meta[config.GROUP_BASENAME]
        roi_basenames = roi_basenames.drop_duplicates().values
        re_ome = imgconf[config.IMAGE_OME_META_REGEXP]
        roi_meta = lib.map_group_re(roi_basenames, re_ome)
        roi_meta[config.GROUP_BASENAME] = roi_basenames
        colmap = {
            imgconf[g]: v
            for g, v in [
                (config.GROUP_SLIDEAC, db.slideacs.slideac_name.key),
                (config.GROUP_PANORMAID, db.sites.site_mcd_panoramaid.key),
                (config.GROUP_ACID, db.acquisitions.acquisition_mcd_acid.key),
                (config.GROUP_ROIID, db.acquisitions.acquisition_mcd_roiid.key),
            ]
        }
        roi_meta = roi_meta.rename(columns=colmap)
        # set default values
        roi_meta = roi_meta.reindex(
            columns=list(colmap.values()) + [config.GROUP_BASENAME], fill_value=np.NAN
        )
        roi_meta[db.acquisitions.acquisition_id.key] = self._query_new_ids(
            db.acquisitions.acquisition_id, roi_meta.shape[0]
        )
        img_meta = pd.merge(
            img_meta,
            roi_meta.loc[
                :, [db.acquisitions.acquisition_id.key, config.GROUP_BASENAME]
            ],
            on=config.GROUP_BASENAME,
        )
        return img_meta, roi_meta

    def _generate_site_table(self, roi_meta):
        idvars = [db.slideacs.slideac_name.key, db.sites.site_mcd_panoramaid.key]
        site_meta = roi_meta.loc[:, idvars]
        site_meta = site_meta.drop_duplicates()

        site_meta[db.sites.site_id.key] = self._query_new_ids(
            db.sites.site_id, site_meta.shape[0]
        )
        roi_meta = pd.merge(
            roi_meta, site_meta.loc[:, idvars + [db.sites.site_id.key]], on=idvars
        )
        return roi_meta, site_meta

    def _generate_slideac_table(self, site_meta):
        slideacs = site_meta[db.slideacs.slideac_name.key]
        slideacs = slideacs.drop_duplicates().values
        imgconf = self.conf[config.CPOUTPUT][config.IMAGES_CSV]
        re_slide = imgconf[config.IMAGE_SLIDE_REGEXP]
        slideac_meta = lib.map_group_re(slideacs, re_slide)
        colmap = {
            imgconf[g]: v
            for g, v in [
                (config.GROUP_SLIDENUMBER, db.slides.slide_number.key),
                (config.GROUP_SAMPLEBLOCKNAME, db.sampleblocks.sampleblock_name.key),
            ]
        }
        slideac_meta = slideac_meta.rename(columns=colmap)
        # set default values
        slideac_meta = slideac_meta.reindex(
            columns=list(colmap.values()), fill_value=np.NAN
        )
        slideac_meta[db.slideacs.slideac_name.key] = slideacs

        slideac_meta[db.slideacs.slideac_id.key] = self._query_new_ids(
            db.slideacs.slideac_id, slideac_meta.shape[0]
        )
        site_meta = pd.merge(
            site_meta,
            slideac_meta.loc[
                :, [db.slideacs.slideac_id.key, db.slideacs.slideac_name.key]
            ],
            on=db.slideacs.slideac_name.key,
        )
        return site_meta, slideac_meta

    def _generate_slide_table(self, slideac_meta):
        slide_meta = slideac_meta.loc[
            :, [db.slides.slide_number.key, db.sampleblocks.sampleblock_name.key]
        ]
        slide_meta = slide_meta.drop_duplicates()
        slide_meta[db.slides.slide_id.key] = self._query_new_ids(
            db.slides.slide_id, slide_meta.shape[0]
        )
        slideac_meta = pd.merge(
            slideac_meta,
            slide_meta.loc[
                :,
                [
                    db.slides.slide_number.key,
                    db.slides.slide_id.key,
                    db.sampleblocks.sampleblock_name.key,
                ],
            ],
            on=[db.slides.slide_number.key, db.sampleblocks.sampleblock_name.key],
        )
        return slideac_meta, slide_meta

    def _generate_sampleblock_table(self, slide_meta):
        sampleblock_meta = slide_meta.loc[:, [db.sampleblocks.sampleblock_name.key]]
        sampleblock_meta = sampleblock_meta.drop_duplicates()
        sampleblock_meta[db.sampleblocks.sampleblock_id.key] = self._query_new_ids(
            db.sampleblocks.sampleblock_id, sampleblock_meta.shape[0]
        )
        slide_meta = pd.merge(
            slide_meta,
            sampleblock_meta.loc[
                :,
                [
                    db.sampleblocks.sampleblock_name.key,
                    db.sampleblocks.sampleblock_id.key,
                ],
            ],
            on=db.sampleblocks.sampleblock_name.key,
        )
        return slide_meta, sampleblock_meta

    def _generate_masks(self):
        cpconf = self.conf[config.CPOUTPUT]
        objects = cpconf[config.MEASUREMENT_CSV][config.OBJECTS]
        prefix = cpconf[config.IMAGES_CSV][config.MASK_FILENAME_PREFIX]
        dat_mask = {
            obj: self._images_csv[[db.images.image_number.key, prefix + obj]].rename(
                columns={prefix + obj: db.masks.mask_filename.key}
            )
            for obj in objects
        }
        dat_mask = pd.concat(dat_mask, names=[db.objects.object_type.key, "idx"])
        dat_mask = dat_mask.reset_index(level=db.objects.object_type.key, drop=False)
        dat_mask = dat_mask.reset_index(drop=True)
        #    if all(dat_mask[db.masks.shape_w.key].isnull()):
        #        """
        #        If the width and height are not in the regexp, load all the
        #        mask and check the width
        #        """
        #        cpconf = self.config[config.CPOUTPUT]
        #        basedir = cpconf[config.IMAGES_CSV][config.MASK_DIR]
        #        if basedir is None:
        #            basedir = self.config[config.CP_DIR]
        #        dat_mask[db.masks.shape_w.key], dat_mask[db.masks.shape_h.key] = \
        #                zip(*dat_mask[db.masks.file_name.key].map(lambda fn:
        #                        tif.imread(os.path.join(basedir, fn)).shape))
        img_dict = {
            n: i
            for n, i in self.main_session.query(
                db.images.image_number, db.images.image_id
            )
        }
        dat_mask[db.masks.image_id.key] = dat_mask[db.images.image_number.key].replace(
            img_dict
        )
        return dat_mask

    def _generate_image_stacks(self):
        stack_col = "column_stack"
        cpconf = self.conf[config.CPOUTPUT]
        dat_stacks = self.bro.doquery(self.main_session.query(db.stacks))
        dat_img = self._images_csv
        prefix = cpconf[config.IMAGES_CSV][config.STACKIMG_FILENAME_PREFIX]
        dat_stacks[stack_col] = dat_stacks[db.stacks.stack_name.key].map(
            lambda x: prefix + x
        )
        fil = dat_stacks[stack_col].isin(dat_img.columns)
        dat_stacks = dat_stacks.loc[fil, :]
        dat_image_stack = dat_img.melt(
            id_vars=[db.images.image_number.key],
            value_vars=dat_stacks[stack_col],
            value_name=db.image_stacks.image_stack_filename.key,
            var_name=stack_col,
        )
        dat_image_stack = dat_image_stack.merge(dat_stacks)
        img_dict = {
            n: i
            for n, i in self.main_session.query(
                db.images.image_number, db.images.image_id
            )
        }
        dat_image_stack[db.image_stacks.image_id.key] = dat_image_stack[
            db.images.image_number.key
        ].replace(img_dict)
        return dat_image_stack

    def _write_masks_table(self):
        masks = self._generate_masks()
        self._bulkinsert(masks, db.masks)

    def _write_image_stacks_table(self):
        img_stacks = self._generate_image_stacks()
        self._bulkinsert(img_stacks, db.image_stacks)

    def _write_objects_table(self):
        """
        Generates and save the cell table
        """
        objects = self._generate_objects()
        self._bulkinsert(objects, db.objects)
        self._bulkinsert(objects, db.valid_objects)

    def _generate_objects(self):
        """
        Genertes the cell table
        """
        objects = pd.DataFrame(
            self._measurement_csv[
                [
                    db.objects.object_number.key,
                    db.objects.object_type.key,
                    db.images.image_number.key,
                ]
            ]
        )

        objects[db.objects.object_id.key] = self._query_new_ids(
            db.objects.object_id, (objects.shape[0])
        )
        # TODO: fix this
        img_dict = {
            n: i
            for n, i in self.main_session.query(
                db.images.image_number, db.images.image_id
            )
        }
        objects[db.images.image_id.key] = objects[db.images.image_number.key].replace(
            img_dict
        )
        return objects

    def _write_measurement_table(self, minimal):
        """
        Generates the Measurement, MeasurementType and MeasurementName
        tables and writes them to an anndata object.

        """
        if self.conf[config.BACKEND] == config.CON_SQLITE:
            for obj_type, meas in self._generate_anndata_measurements():
                ioan = io_anndata.IoAnnData(self.bro, obj_type)
                ioan.initialize_anndata(meas)

    def _register_measurement_meta(self, dat_meas):
        meas_cols = list(
            set(dat_meas.columns)
            - {
                db.objects.object_type.key,
                db.images.image_number.key,
                db.objects.object_number.key,
                db.objects.object_id.key,
            }
        )
        meta = pd.Series(meas_cols).apply(
            lambda x: lib.find_measurementmeta(
                self._stacks,
                x,
                no_stack_str=OBJECTS_STACKNAME,
                no_plane_string=OBJECTS_PLANEID,
            )
        )
        meta.columns = [
            "variable",
            db.measurement_types.measurement_type.key,
            db.measurement_names.measurement_name.key,
            db.stacks.stack_name.key,
            db.ref_planes.ref_plane_number.key,
        ]
        meta = meta.loc[meta["variable"] != "", :]
        meta[db.ref_planes.ref_plane_number.key] = meta[
            db.ref_planes.ref_plane_number.key
        ].map(lambda x: int(x.replace("c", "")))

        dat_planeids = pd.read_sql(
            self.main_session.query(
                db.stacks.stack_name, db.planes.ref_plane_number, db.planes.plane_id
            )
            .join(db.planes)
            .statement,
            self.db_conn,
        )

        meta = meta.merge(dat_planeids)
        meta = self.bro.processing.measurement_maker.register_measurements(meta)
        return meta

    def _register_objects(self, dat_meas):
        """
        registers the objects
        """
        obj_metavars = [
            db.objects.object_number.key,
            db.objects.object_type.key,
            db.images.image_number.key,
        ]
        dat_objmeta = dat_meas.loc[:, obj_metavars]
        dat_objmeta = self.bro.processing.measurement_maker.register_objects(
            dat_objmeta, assume_new=True
        )
        return dat_objmeta

    def _generate_anndata_measurements(self):
        """
        Loads the measurements and converts them to the anndata format
        ->

        """
        conf_meas = self.conf[config.CPOUTPUT][config.MEASUREMENT_CSV]
        for obj_type in conf_meas[config.OBJECTS]:
            logging.debug(f"Read {obj_type}:")
            dat_meas = next(self._read_objtype_measurements(obj_type, chunksize=None))
            # register the measurements
            logging.debug("Register measurements:")
            dat_measmeta = self._register_measurement_meta(dat_meas)
            logging.debug("Register objects:")
            # register the objects
            # -> This adds objectid to the table
            dat_objmeta = self._register_objects(dat_meas)
            dat_meas = dat_meas.merge(dat_objmeta)
            # set the object_id as index
            dat_meas.set_index(db.objects.object_id.key, inplace=True)
            dat_meas.drop(columns=dat_objmeta.columns, errors="ignore", inplace=True)
            variables = dat_measmeta["variable"]
            dat_meas = dat_meas.loc[:, variables].rename(
                columns={
                    v: int(i)
                    for v, i in zip(
                        dat_measmeta["variable"],
                        dat_measmeta[db.measurements.measurement_id.key],
                    )
                }
            )
            yield obj_type, dat_meas

    def _generate_measurements(self, minimal, chuncksize=3000, longform=True):
        conf_meas = self.conf[config.CPOUTPUT][config.MEASUREMENT_CSV]
        for obj_type in conf_meas[config.OBJECTS]:
            logging.debug(f"Read {obj_type}:")
            for dat_meas in self._read_objtype_measurements(obj_type, chuncksize):
                # register the measurements
                logging.debug("Register measurements:")
                measurement_meta = self._register_measurement_meta(dat_meas)
                logging.debug("Register objects:")
                # register the objects
                object_meta = self._register_objects(dat_meas)
                logging.debug("Reshape values:")
                # get the measurement data
                variables = measurement_meta["variable"]
                dat_measvals = dat_meas.loc[:, variables]
                value_col = dat_measvals.values.flatten(order="C")
                meas_id = measurement_meta[
                    db.object_measurements.measurement_id.key
                ].values
                meas_id = np.tile(meas_id, dat_meas.shape[0])
                obj_id = object_meta[db.object_measurements.object_id.key].values
                obj_id = np.repeat(obj_id, len(variables))
                logging.debug("Construct dataframe")
                measurements = pd.DataFrame(
                    {
                        db.object_measurements.object_id.key: obj_id,
                        db.object_measurements.measurement_id.key: meas_id,
                        db.object_measurements.value.key: value_col,
                    }
                )
                logging.debug("Replace non finite")
                measurements[db.object_measurements.value.key].replace(
                    np.inf, 2 ** 16, inplace=True
                )
                measurements[db.object_measurements.value.key].replace(
                    -np.inf, -(2 ** 16), inplace=True
                )
                measurements.dropna(inplace=True)

                logging.debug("Start uploading")
                yield measurements

    def _generate_object_relation_types(self):
        dat_relations = self._relation_csv
        dat_types = pd.DataFrame(
            dat_relations.loc[:, db.object_relation_types.object_relationtype_name.key]
        ).drop_duplicates()
        dat_types[
            db.object_relation_types.object_relationtype_id.key
        ] = self._query_new_ids(
            db.object_relation_types.object_relationtype_id, (dat_types.shape[0])
        )
        return dat_types

    def _generate_object_relations(self):
        logging.debug("Start img_dict")
        img_dict = {
            n: i
            for n, i in self.main_session.query(
                db.images.image_number, db.images.image_id
            )
        }
        logging.debug("Start relation_dict")
        relation_dict = {
            n: i
            for n, i in self.main_session.query(
                db.object_relation_types.object_relationtype_name,
                db.object_relation_types.object_relationtype_id,
            )
        }
        logging.debug("Start obj dict")
        obj_dict = {
            (imgid, objnr, objtype): objid
            for imgid, objnr, objtype, objid in self.main_session.query(
                db.objects.image_id,
                db.objects.object_number,
                db.objects.object_type,
                db.objects.object_id,
            ).all()
        }
        logging.debug("End obj dict")

        dat_relations = self._relation_csv
        logging.debug("Start replacing objfrom")
        dat_relations["timg"] = dat_relations[config.IMAGENUMBER_FROM].replace(img_dict)
        dat_relations[db.object_relations.object_id_parent.key] = dat_relations.loc[
            :, ["timg", config.OBJECTNUMBER_FROM, config.OBJECTTYPE_FROM]
        ].apply(lambda x: obj_dict.get((x[0], x[1], x[2])), axis=1)
        logging.debug("Start replacing ids objto")
        dat_relations["timg"] = dat_relations[config.IMAGENUMBER_TO].replace(img_dict)
        dat_relations[db.object_relations.object_id_child.key] = dat_relations.loc[
            :, ["timg", config.OBJECTNUMBER_TO, config.OBJECTTYPE_TO]
        ].apply(lambda x: obj_dict.get((x[0], x[1], x[2])), axis=1)
        dat_relations[db.object_relations.object_relationtype_id.key] = dat_relations[
            db.object_relation_types.object_relationtype_name.key
        ].replace(relation_dict)
        return dat_relations

    def _write_object_relations_table(self):
        logging.debug("start generate object_relation_types")
        relation_types = self._generate_object_relation_types()
        self._bulkinsert(relation_types, db.object_relation_types)
        logging.debug("start generate object_relations")
        relations = self._generate_object_relations()
        self._bulkinsert(relations, db.object_relations)

    def _write_pannel_table(self):
        pannel = self._generate_pannel_table()
        self._bulkinsert(pannel, db.pannel)

    def _generate_pannel_table(self):
        csv_pannel = self.pannel
        conf_pannel = self.conf[config.PANNEL_CSV]
        col_map = {
            conf_pannel[c]: target
            for c, target in [
                (config.PANEL_CSV_CHANNEL_NAME, db.pannel.metal.key),
                (config.PANEL_CSV_ILASTIK_NAME, db.pannel.is_ilastik.key),
                (config.PANEL_CSV_BARCODE_NAME, db.pannel.is_barcode.key),
                (config.PANEL_CSV_CLONE_NAME, db.pannel.antibody_clone.key),
                (config.PANEL_CSV_CONCENTRATION_NAME, db.pannel.concentration.key),
                (config.PANEL_CSV_TARGET_NAME, db.pannel.target.key),
                (config.PANEL_CSV_TUBE_NAME, db.pannel.tube_number.key),
            ]
        }
        cols = [c for c in col_map]
        csv_pannel.drop(list(set(csv_pannel.columns) - set(cols)), axis=1, inplace=True)
        csv_pannel = csv_pannel.rename(columns=col_map)
        # correct conc to Float
        csv_pannel[db.pannel.concentration.key] = csv_pannel[
            db.pannel.concentration.key
        ].apply(lambda x: float(re.findall(r"[-+]?\d*\.\d+|\d+", x)[0]))
        # correct boolean to logical
        csv_pannel.loc[:, [db.pannel.is_barcode.key, db.pannel.is_ilastik.key]] = (
            csv_pannel.loc[:, [db.pannel.is_barcode.key, db.pannel.is_ilastik.key]] == 1
        )
        return csv_pannel

    def _write_condition_table(self):
        conditions = self._generate_condition_table()
        if conditions is not None:
            self._bulkinsert(conditions, db.conditions)

    def _generate_condition_table(self):
        """
        Generates the condition metadata table based on a barcode and or a condition file
        """
        conf_layout = self.conf[config.LAYOUT_CSV]
        exp_layout = self.experiment_layout
        barcode_key = self.barcode_key
        if (exp_layout is None) and (barcode_key is None):
            return None
        if exp_layout is None:
            data = barcode_key
        elif barcode_key is None:
            data = exp_layout
        else:
            data = exp_layout.merge(barcode_key)
        data = lib.fill_null(data, db.conditions)

        # Legacy: split barcode in x and y
        def get_y(x):
            return x[0]

        def get_x(x):
            return int(x[1:])

        data[db.conditions.bc_x.key] = data[db.conditions.well_name.key].map(get_x)
        data[db.conditions.bc_y.key] = data[db.conditions.well_name.key].map(get_y)

        # Assign the condition IDs:
        ncond = data.shape[0]
        data[db.conditions.condition_id.key] = self._query_new_ids(
            db.conditions.condition_id, ncond
        )

        # Add the sampleblock id:
        sample_dict = {
            n: i
            for n, i in self.main_session.query(
                db.sampleblocks.sampleblock_name, db.sampleblocks.sampleblock_id
            )
        }
        if db.sampleblocks.sampleblock_name.key in data:
            # If sampleblocks are used, replace them by sample ids
            data[db.sampleblocks.sampleblock_id.key] = data[
                db.sampleblocks.sampleblock_name.key
            ].replace(sample_dict)
        elif len(sample_dict) == 1:
            # If there is only one sample, assume this is the one that the conditions refer to
            data[db.sampleblocks.sampleblock_id.key] = list(sample_dict.values())[0]
        else:
            raise (
                """Sampleblocks are used in the `slide_regexp`, but not refered to in the experiment layout!\n
                    Either remove them in the regexp or add them to the layout!"""
            )
        return data

    def _query_new_ids(self, id_col, n):
        """
        Queries non used id's from the database
        Args:
            id_col: a sqlalchemy column object corresponding
                to a column in a table
            n: how many id's are requested
        """
        if self.conf[config.BACKEND] == config.CON_POSTGRESQL:
            session = self.session_maker()
            str_seq = str(id_col).replace(".", "_") + "_seq"
            # this will instantiate the sequence if it is not
            # done so yet
            # otherwise it will lead to a gap of 1 between the ids
            # which should not matter
            _ = session.execute(sa.schema.Sequence(str_seq))

            session.execute("ALTER SEQUENCE " + str_seq + " INCREMENT " + str(n))
            i = session.execute(sa.schema.Sequence(str_seq))
            session.execute("ALTER SEQUENCE " + str_seq + " INCREMENT " + str(1))
            session.commit()
            return range(i - n + 1, i + 1)
        else:
            # self.main_session.query(func
            prev_max = self.main_session.query(sa.func.max(id_col)).scalar()
            if prev_max is None:
                prev_max = 0
            return range(prev_max + 1, prev_max + n + 1)

    #########################################################################
    #########################################################################
    #                           setter functions:                           #
    #########################################################################
    #########################################################################

    def _bulkinsert(self, data, table, drop=None):
        """_bulkinsert
        This function is used for Bulk inserting data to the database.
        Note that dropping all entries in a table can fail because of
        foregn key constraints. It is recommended to only use this method
        at the first data import.

        Args:
            DataFrame data: the data to be inserted
            sqlalchemy table: the target table
            bool drop: if the table should be emptied before inserting.
                default False.
        """
        if drop is None:
            drop = False

        dbtable = table.__table__.name
        if drop:
            session = self.main_session
            session.query(table).delete()
            session.commit()

        logging.debug("Insert table of dimension: " + str(data.shape))
        data = self._clean_columns(data, table)
        data.to_sql(
            dbtable,
            self.db_conn,
            if_exists="append",
            index=False,
            method="multi",
            chunksize=999,
        )
        # odo(data, dbtable)
        self.main_session.commit()

    def _clean_columns(self, data, table):
        """
        Removes columns not in table, adds columns with default value None if they are missing from data.
        Args:
            data: a pandas data table
            table: an sqlalchemy table definition

        Returns:
            The cleaned pandas data table

        """
        data_cols = data.columns
        table_cols = table.__table__.columns.keys()
        data = data.reindex(columns=table_cols, fill_value=None)
        return data

    def _add_generic_tuple(
        self, data, table, query=None, replace=False, backup=False, pg=False
    ):
        """add_generic_tuple
        adds tuples from date to the database and returns non stored or
        deleted values.

        Args:
            Pandas DataFrame data: dataframe containing the data. It is
                required to name the columns according to the db schema
            Sqlalchemy Table table: the table object to be added to
            sqlalchemy query query: query object to retrieve existing tuples.
                best option: query for all keys! If no query is supplied,
                a query will be generated based on the table keys
            bool replace: if existing tuples should be replaced
            backup: only used if replace = True. Specifies whether a table
                with the deleted tuples should be returned. Can speed up
                operation

        Returns:
            Pandas.DataFrame containing the deleted tuples. These can be used
                to restore the old ones.
            Pandas.DataFrame containing the unstored rows

        """
        data = data.reset_index(drop=True)
        key_cols = [key.name for key in inspect(table).primary_key]
        if query is None:
            query = self.main_session.query(table)
            for key in key_cols:
                filt_in = data[key].astype(str).unique()
                query = query.filter(table.__table__.columns[key].in_(filt_in))
        if replace:
            if backup:
                backup = pd.read_sql(query.statement, self.db_conn)
            else:
                backup = None

            query.delete(synchronize_session="fetch")
            self.main_session.commit()
            self._bulkinsert(data, table)

            return backup, None
        else:
            backup = pd.read_sql(query.statement, self.db_conn)
            current = backup.copy()
            if current.shape[0] == 0:
                # if nothing already in the database (=current empty), store everything
                storable = data
                # unstored will be an empty dataframe
                unstored = current
            else:
                zw = (
                    data[key_cols].append(current[key_cols]).drop_duplicates(keep=False)
                )
                storable = data.merge(zw)
                unstored = data.merge(zw, how="outer")

            lm, ls = len(data), len(storable)
            if lm != ls:
                miss = lm - ls
                stri = "There were "
                stri += str(miss)
                stri += " rows that were not updated in "
                stri += table.__tablename__
                stri += "! This does not mean that something went wrong, but "
                stri += "maybe you tried to readd some rows."
                warnings.warn(stri, UserWarning)

            self._bulkinsert(storable, table)

            return None, unstored

    def reset_valid_images(self):
        sel = sa.select([db.images.image_id]).where(
            ~db.images.image_id.in_(self.main_session.query(db.valid_images.image_id))
        )
        ins = sa.insert(db.valid_images).from_select(
            [db.valid_images.image_id.key], sel
        )
        self.main_session.execute(ins)
        self.main_session.commit()

    def reset_valid_objects(self):
        sel = sa.select([db.objects.object_id]).where(
            ~db.objects.object_id.in_(
                self.main_session.query(db.valid_objects.object_id)
            )
        )
        ins = sa.insert(db.valid_objects).from_select(
            [db.valid_objects.object_id.key], sel
        )
        self.main_session.execute(ins)
        self.main_session.commit()

    #########################################################################
    #########################################################################
    #                           getter functions:                           #
    #########################################################################
    #########################################################################
    def get_measmeta_query(self, session=None):
        """
        Returns a measurement object that queries table with the most important
        information do identify a measurement.
        """
        if session is None:
            session = self.main_session

        query = (
            session.query(db.measurements)
            .join(db.measurement_types)
            .join(db.measurement_names)
            .join(db.planes)
            .join(db.stacks)
            .join(db.ref_planes)
            .join(db.ref_stacks)
        )
        return query

    def get_objectmeta_query(self, session=None, valid_objects=True, valid_images=True):
        """
        Returns a query object that queries table with the most important
        information do identify an object.
        """
        if session is None:
            session = self.main_session

        query = (
            session.query(db.objects)
            .join(db.images)
            .join(db.acquisitions)
            .join(db.sites)
            .join(db.slideacs)
            .join(db.slides)
            .join(db.sampleblocks)
        )

        if valid_objects:
            query = query.join(db.valid_objects)
        if valid_images:
            query = query.join(db.valid_images)
        return query

    def _get_namekey_dict(self, namecol, idcol, names):
        """
        Generates a name: idcol dictionary from a table
        while filtering for names in the namecol

        namecol: A sql column, e.g. db.Stack.StackName
        idcol: A sql id column, e.g. db.Stack.StackID
        names: names to be queried

        """
        d = {
            n: i
            for n, i in (
                self.main_session.query(namecol, idcol).filter(namecol.in_(names))
            )
        }
        return d

    # Properties:
    @property
    def pannel(self):
        if self._pannel is None:
            self._read_pannel()
        return self._pannel

    @property
    def _name_dict(self):
        conf_pannel = self.conf[config.PANNEL_CSV]
        col_channel = conf_pannel[config.CHANNEL_NAME]
        col_name = conf_pannel[config.DISPLAY_NAME]
        name_dict = {
            metal: name
            for metal, name in zip(self._pannel[col_channel], self._pannel[col_name])
        }
        return name_dict

    @property
    def _stacks(self):
        stacks = list(
            self._stack_relation_csv[self.conf[config.STACK_RELATIONS][config.STACK]]
        )
        stacks += [s for s in [st for st in self.stack_csvs]]
        return set(stacks)

    @property
    def session_maker(self):
        """
        Returns the session maker object for the current database connection
        """
        if self._session_maker is None:
            self._session_maker = sessionmaker(bind=self.db_conn)
        return self._session_maker

    @property
    def main_session(self):
        """
        Returns the current database main session
        to query the database in an orm way.
        """
        if self._session is None:
            self._session = self.session_maker()
        return self._session

    def query_df(self, query: sa.orm.query.Query) -> pd.DataFrame:
        """
        Executes an sqlalchemy query and returns the corresponding dataframe.

        Args:
            query: an sqlalchemy query
        Returns:
            The resulting dataframe

        """
        return pd.read_sql(query.statement, self.db_conn)
